/* partition.c */
/* this program is an implementation of the clustering flow-chart 
   in the paper "Node-Tearing Nodal Analysis" Memo # ERL-M582 
   this program has been customized to partition a graph for 
   a Gauss-Seidel-based parallel solver.  */

#include <stdio.h>

#include <X11/Xlib.h>    /* X-image */
#include <X11/Xutil.h>   /* X-image */
#include <X11/Xos.h>     /* X-image */
#include <math.h>        /* X-image */
#include <sys/file.h>    /* X-image */
#include "icon_bitmap"   /* X-image */

#define INFI    1073741824
#define YES              1
#define NO               0
#define FOREVER          1
#define BEFORE_COLORING  0
#define AFTER_COLORING   1

#include "sets_pt.h"
#include "bdb_struct_gs_part.h"

/* define a grayscale bitmap that represents the sparsity pattern
   in the matrix 
    nonzero -> black=1
    zero    -> white=254*/

#define NONZERO        1 /* black */
#define ZERO         248 /* white */
#define COLOR_1        1
#define COLOR_2       20
#define COLOR_3       40
#define COLOR_4       60
#define COLOR_5       80
#define COLOR_6      100
#define COLOR_7      120
#define COLOR_8      140
#define COLOR_9      160
#define COLOR_10     180

/* values for window_size in main, is window big enough to be useful */
#define SMALL        1   /* X-image */
#define OK           0   /* X-image */
#define MAX_COLORS 256   /* X-image */ 

/* The calculation of data clusters will be performed on the data pairs
   representation of the network.  This will simplify the network
   reordering - nodes just need to be renumbered.  */

typedef struct Node{int NodeNum;
                    int NodeInUse;} Node;

typedef struct AdjList{Node l;
                       Node r;} AdjList;

AdjList *adjlist;   /* adjlist[]  */
AdjList *adjlist1;  /* adjlist1[] */
int nadjlist;       /* number of adjacency pairs */

typedef struct NodeLB{int NodeNum;
                      int NodeNumR;
                      int Color;
                      int ProcNum;
                      Set *Neighbors;
                      Set *ColoredNeighbors;
                      Set *ColorsOfNeighbors;} NodeLB;
NodeLB *nodeLB; /* nodes in the last diagonal block */
NodeLB *nodeLBT;

typedef struct AdjListLB{NodeLB l;
                         NodeLB r;} AdjListLB;

AdjListLB *adjlistLB;   /* adjlistLB[]  */
int nadjlistLB;       /* number of adjacency pairs in the last diagonal block */

typedef struct NodeListObj{int NodeNum;
                           int NodeInUse;} NodeList;

NodeList *nodelist; /* nodelist[] */

Set    *N2;           /* the set N2 */
SoSets *N1;           /* the set of sets of N1a... or the set of clusters */
SoSets *rN1;          /* the set of sets of N1a... recursively ordered set of clusters*/
SoSets *sN1;          /* the set of sets of N1a... ordered according to the super clusters */

/* The tableau is stored in stored as an array of structures -
   (1) the node number and (2) the degree of the node */

typedef struct Tabl{int NodeNum;
                    Set *AdjacentSet;
                    int ContourNum;} Tableau;

Tableau *tableau;   /* tableau[] */
int ntab;           /* number of elements in the tableau */

typedef struct SparseMatrixSet{Set *Row;} SparseMatrixSet;
SparseMatrixSet *SparseMatrix; /* SparseMatrix[] */

typedef struct IndBlocks{Set *N1i;
                         int blockstart;
                         int blockend;} IndBlocks;
IndBlocks *IndBlk; /* IndBlk[] */

int n;               /* the number of columns/rows in A */
int ndisplay;        /* the number of columns/rows in the display */
int ncomb;           /* the number of pixels to be combined to condense the display */
int ntot;            /* the number of nonzeros in A */
int nOpsLastBlockGS; /* number of operations in the last block */

int nmax;            /* maximum size of a cluster */
int nmin;            /* minimum size of a cluster */

/* define the structure to perform the pigeon-hole load balancing */

typedef struct LoadBalance{int    nOpsBlockGS;
                           int    nOpsBorderGS;
                           int    nOpsTotalGS;
                           int    nOpsTotal;
                           SoSets *SuperCluster;} LoadBalance;
LoadBalance *loadbal;   /* loadbal[] */

short **bitmap;      /* bitmap[][] display bitmap */ 
int xdisplay;
/* Display and screen are used as arguments to nearly every Xlib
 * routine, so it simplifies routine calls to declare them global.
 */
Display      *display;          /* X-image */
int          screen;            /* X-image */
XImage       *picture;          /* X-image */
unsigned int width, height;     /* X-image */
int    gargc;                   /* X-image */
char **gargv;                   /* X-image */
/*---------------------------------------------------------*/
Proc            *proc;          /* proc[]          */
Block           *blocks;        /* blocks[]        */
Diag            *diag;          /* diag[]          */
BlockRow        *blockrow;      /* blockrow[]      */
LowerBorder     *lborder;       /* lborder[]       */
LBorderRow      *lborderrow;    /* lborderrow[]    */
LastBlockColor **lbcolor;       /* lbcolor[][]     */
LastBlockColor  *lbcolortmp;    /* lbcolortmp[]    */
LastDiag        *ldiag;         /* ldiag[]         */
LastBlockRow    *lblockrow;     /* lblockrow[]     */
/*---------------------------------------------------------*/
int            nprocs;      /* # of procs                */
int            nblocks;     /* # of blocks               */
int            ndiag;       /* dimension of diag         */
int            nblockrow;   /* dimension of blockrow     */
int            nlborder;    /* dimension of lborder      */
int            nlborderrow; /* dimension of lborderrow   */
int            ncolors;     /* number of unique colors   */
int            nldiag;      /* dimension of ldiag        */
int            nlblockrow;  /* dimension of lblockrow    */
/*---------------------------------------------------------*/
CommLB         **commlb;
CommLB          *commlbtmp;
/*---------------------------------------------------------*/
/*  size of message to update last block for each processor --- 
different data structure for each proc */

int **ulbbufsize;     /* ulbbufsize[iproc][jproc] */ 
int  *ulbbufsizetmp;
int *ReorderLst;  /* ReorderLst[] */
/*---------------------------------------------------------*/
int    ln;    /* dimension of the last block */
/*---------------------------------------------------------*/
void main0();
void main1();
void main2();
void main3();
void main4();
char *MakeOutputFilename();
char *MakeOutputFilenameLB();
void InputNetwork();
void ResetAdjList();
void IdClusters();
void SubClusterBlocks();
void ResetNodesInN1i();
void IdSubClusters();
char *MakeClusterName();
char *MakeSubClusterName();
void InitNodeList();
void InitTableau();
int  StartNode();
void UpdateTableau();
int  SelectNextNodeInTableau();
void UpdateAdjacencyList();
void ResetAdjacencyList();
void ReorderNetworkClusters();
void ColorLB();
void MinDegreeReorder();
void NonZeroDensityInLastBlock();
void GenerateBitmap();
void GenerateBitmapLB();
void PrintL();
void PrintLB();
void CountOpsBDB();
void PigeonHoleLoadBalance();
void BDBFormData();
void OutputBDBFormData();

double drand48();
void   srand48();

void Ximage();         /* X-image */
void set_gray_scale(); /* X-image */
void get_GC();         /* X-image */
int  load_data();      /* X-image */

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
main(argc,argv)
  int argc;
  char **argv;
{
  int i,j,opt;
  int seed = 10770;
  int nprocsmin,nprocsmax;
  char filename[40];

  if (argc < 10)
   {
    printf("Usage: (1) matrix in adjacency list form \n");
    printf("       (2) nmax - maximum number of nodes in a cluster\n");
    printf("       (3) nmin - suggested minimum number of nodes in a cluster\n");
    printf("       (4) nprocsmin - min number of procs for load balancing\n");
    printf("       (5) nprocsmax - max number of procs for load balancing\n");
    printf("       (6) bitmap output name key (none)\n");
    printf("       (7) BDB-form matrix output name \n");
    printf("       (8) b vector output name \n");
    printf("       (9) options for xdisplay \n");
    printf("           0 - provide no xdisplay \n");
    printf("           1 - provide an xdisplay \n");
    printf("      (10) options when running the program \n");
    printf("           1 - data in original ordering \n");
    printf("           2 - data after clustering and reordering \n");
    printf("           3 - data after subclustering and reordering \n");
    exit(1);
   } /* end if */

/* save the input string */
  gargc = argc;
  gargv = argv;

 printf("matrix in adjacency list form                     - argv[1]=%s\n",argv[1]);
 printf("maximum number of nodes in a cluster              - argv[2]=%s\n",argv[2]);
 printf("minimum number of nodes in a cluster              - argv[3]=%s\n",argv[3]);
 printf("minimum number of processors                      - argv[4]=%s\n",argv[4]);
 printf("maximum number of processors                      - argv[5]=%s\n",argv[5]);
 printf("bitmap output name key (none to suppress output)  - argv[6]=%s\n",argv[6]);
 printf("BDB-form matrix output name                       - argv[7]=%s\n",argv[7]);
 printf("b vector output name                              - argv[8]=%s\n",argv[8]);
 printf("option for an X-dosplay of the graph              - argv[9]=%s\n",argv[9]);
 printf("option when running the program                  - argv[10]=%s\n",argv[10]);

 sscanf(argv[2],"%d",&nmax);
 printf("maximum number of nodes in a cluster - nmax=%d \n",nmax);
 sscanf(argv[3],"%d",&nmin);
 printf("minimum number of nodes in a cluster - nmin=%d \n",nmin);
 sscanf(argv[4],"%d",&nprocsmin);
 printf("minimum number of processors - nprocsmin=%d \n",nprocsmin);
 nprocs = nprocsmin;
 sscanf(argv[5],"%d",&nprocsmax);
 printf("maximum number of processors - nprocsmax=%d \n",nprocsmax);
 sscanf(argv[9],"%d",&xdisplay);
 printf("option for xdisplay - xdisplay=%d \n",xdisplay);
 sscanf(argv[10],"%d",&opt);
 printf("option when running the program  - opt=%d \n",opt);
 fflush(stdout);

/* initialize the SetElementPool */
  SetElementPool = NULL;
  SetElementPool = InitSet(SetElementPool,"SetElementPool");
  SetOfSetsElementPool = NULL;
  SetOfSetsElementPool = InitSetOfSets(SetOfSetsElementPool,"SetOfSetsElementPool");
  srand48(seed);
  ReorderLst    = NULL;
  adjlistLB     = NULL;
  nodeLB        = NULL;
  nodeLBT       = NULL;
  IndBlk        = NULL;
  loadbal       = NULL;
  proc          = NULL;
  blocks        = NULL;
  diag          = NULL;
  blockrow      = NULL;
  lborder       = NULL;
  lborderrow    = NULL;
  lbcolor       = NULL;
  lbcolortmp    = NULL;
  ldiag         = NULL;
  lblockrow     = NULL;
  ulbbufsize    = NULL;
  ulbbufsizetmp = NULL;
  commlb        = NULL;
  commlbtmp     = NULL;

printf("InputNetwork \n");
fflush(stdout);
  InputNetwork(argv[1]); /* input pairs of data */

switch (opt)
  {
   case 1: printf("%s data in original ordering \n",argv[6]);
           main0(argv[6]);
           break;

   case 2: printf("%s data after clustering  \n",argv[6]);
           main1(argv[6]);
           printf("%s data after superclustering \n",argv[6]);
           main2(argv[6],argv[7],argv[8],nprocsmin,nprocsmax);
           break;

   case 3: printf("%s data after subclustering and reordering \n",argv[6]);
           main3(argv[6]);
           printf("%s data after superclustering \n",argv[6]);
           main4(argv[6],argv[7],argv[8],nprocsmin,nprocsmax);
           break;

   default: printf("Incorect option entered --- %d \n",opt);
  } /* end switch */

} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main0(namekey)
  char *namekey;
{
  char filename[40];

printf("%s (main0)\n",namekey);

printf("%s data in original ordering (main0) \n",namekey);
printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,0)); 

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the graph */
} /* end main0() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main1(namekey)
  char *namekey;
{
  char filename[40];

printf("%s (main1)\n",namekey);

printf("ResetAdjList \n");
fflush(stdout);
  ResetAdjList();  /* reset the adjacency List to the initial state */

printf("IdClusters \n");
fflush(stdout);
  IdClusters();  /* identify clusters */

printf("ReorderNetworkClusters \n");
fflush(stdout);
  ReorderNetworkClusters(N1,namekey); /* produce block-bordered matrix */

printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,0)); 

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the graph */
} /* end main1() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main2(namekey,FileNameA,FileNameb,nprocsmin,nprocsmax)
  char *namekey,*FileNameA,*FileNameb;
  int nprocsmin,nprocsmax;
{
  char filename[40];
  int i;

printf("%s (main2)\n",namekey);

  printf("CountOpsBDB \n");
  fflush(stdout);
  CountOpsBDB(); 

  i = 0;
  for (nprocs=nprocsmin; nprocs<=nprocsmax; nprocs *= 2)
   {
    printf("\n\nPigeonHoleLoadBalance  --  nprocs=%d\n",nprocs);
    fflush(stdout);
/* calculate the load balance for nprocs processors */
    PigeonHoleLoadBalance(N1,FileNameA,FileNameb,namekey);
   } /* end for nprocs */

} /* end main2() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main3(namekey)
  char *namekey;
{
  char filename[40];

printf("%s (main3)\n",namekey);

printf("ResetAdjList \n");
fflush(stdout);
  ResetAdjList();  /* reset the adjacency List to the initial state */

printf("IdClusters \n");
fflush(stdout);
  IdClusters();  /* identify clusters */

printf("SubClusterBlocks \n");
fflush(stdout);
  SubClusterBlocks();  /* identify subclusters */

printf("ReorderNetworkClusters \n");
fflush(stdout);
  ReorderNetworkClusters(rN1,namekey); /* produce recursive block-bordered matrix */

} /* end main3() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void main4(namekey,FileNameA,FileNameb,nprocsmin,nprocsmax)
  char *namekey,*FileNameA,*FileNameb;
  int nprocsmin,nprocsmax;
{
  char filename[40];
  int i;

printf("%s (main4)\n",namekey);

  i = 0;
  for (nprocs=nprocsmin; nprocs<=nprocsmax; nprocs *= 2)
   {
printf("PigeonHoleLoadBalance  --  nprocs=%d\n",nprocs);
fflush(stdout);
/* calculate the load balance for nprocs processors */
    PigeonHoleLoadBalance(rN1,FileNameA,FileNameb,namekey); 

   } /* end for nprocs */

} /* end main4() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeOutputFilename(key,filename,i)
  char *key,*filename;
  int i;
{
  int j,len;

  if (strcmp(key,"none") == 0)
   {
    strcpy(filename,key);
   } /* end if */
  else
   {
    len = strlen(key);
    for (j=0; j<len; j++)
      filename[j] = key[j];
    filename[j++] = '-';
    filename[j++] = (char) (((nmax%1000)/100) + 48);
    filename[j++] = (char) (((nmax%100)/10) + 48);
    filename[j++] = (char) ((nmax%10) + 48);
    filename[j++] = '-';
    filename[j++] = (char) (((nmin%1000)/100) + 48);
    filename[j++] = (char) (((nmin%100)/10) + 48);
    filename[j++] = (char) ((nmin%10) + 48);
    filename[j++] = '-';
    filename[j++] = (char) (((i%1000)/100) + 48);
    filename[j++] = (char) (((i%100)/10) + 48);
    filename[j++] = (char) ((i%10) + 48);
    filename[j++] = '\0';
   } /* end else */
/*
printf("Output filename for i=%d is |%s| \n",i,filename);
*/

  return filename;
} /* end MakeOutputFilename() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeOutputFilenameLB(namekey,filename,i)
  char *namekey,*filename;
  int i;
{
  int j,k,len;
  char *stringBEFORE = "BEFORE_COLORING";
  char *stringAFTER  = "AFTER_COLORING";

printf("filename key for i=%d is |%s| \n",i,namekey);
fflush(stdout);

  if (strcmp(namekey,"none") == 0)
   {
    strcpy(filename,namekey);
   } /* end if */
  else
   {
    len = strlen(namekey);
printf("\tstrlen(%s)=%d \n",namekey,len);
    for (j=0; j<len; j++)
      filename[j] = namekey[j];
    filename[j++] = '-';
    if (i==BEFORE_COLORING)
      {
       len = strlen(stringBEFORE);
       for (k=0; k<len; k++)
        filename[j+k] = stringBEFORE[k];
       filename[k+j++] = '-';
       filename[k+j++] = (char) (((nprocs%1000)/100) + 48);
       filename[k+j++] = (char) (((nprocs%100)/10) + 48);
       filename[k+j++] = (char) ((nprocs%10) + 48);
       filename[k+j++] = '\0';
      } /* end if */
    else if (i==AFTER_COLORING)
      {
       len = strlen(stringAFTER);
       for (k=0; k<len; k++)
        filename[j+k] = stringAFTER[k];
       filename[k+j++] = '-';
       filename[k+j++] = (char) (((nprocs%1000)/100) + 48);
       filename[k+j++] = (char) (((nprocs%100)/10) + 48);
       filename[k+j++] = (char) ((nprocs%10) + 48);
       filename[k+j++] = '\0';
      } /* end else */
   } /* end else */

printf("Output filename for i=%d is |%s| \n",i,filename);
fflush(stdout);

  return filename;
} /* end MakeOutputFilenameLB() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InputNetwork(filename)
  char *filename;
{
 int i,j,k,l,*histo,*hhisto;
 char *name;
 short *temp;
 FILE *fd;

 printf("read in the sparse matrix\n");
 fflush(stdout);

/* read in the sparse matrix */

printf("filename=|%s| \n",filename);
fflush(stdout);

 fd = fopen(filename,"r");

 fscanf(fd,"%d %d",&n,&nadjlist);

printf("n=|%d| -- nadjlist=|%d| \n",n,nadjlist);
fflush(stdout);

/* allocate memory for the adjacency list */
 adjlist = (AdjList *) calloc((nadjlist+1),sizeof(AdjList));
 if (adjlist == NULL)
  {
   printf("can't allocate memory for adjlist[] \n");
   exit(1);
  } /* end if */
 adjlist1 = (AdjList *) calloc((nadjlist+1),sizeof(AdjList));
 if (adjlist1 == NULL)
  {
   printf("can't allocate memory for adjlist1[] \n");
   exit(1);
  } /* end if */
/* allocate memory for the histogram */
 histo = (int *) calloc((n+1),sizeof(int));
 if (histo == NULL)
  {
   printf("can't allocate memory for histo[] \n");
   exit(1);
  } /* end if */
 for(i=0;i<=n;i++) histo[i]=0;
/* allocate memory for the histogram of histograms*/
 hhisto = (int *) calloc((n+1),sizeof(int));
 if (hhisto == NULL)
  {
   printf("can't allocate memory for hhisto[] \n");
   exit(1);
  } /* end if */
 for(i=0;i<=n;i++) hhisto[i]=0;

 for (k=1;k<=nadjlist;k++)
  {
   fscanf(fd,"%d  %d %d",&l,&i,&j);
/*
printf("  |%d| - (%d,%d)\n",k,i,j);
fflush(stdout);
*/
   adjlist[k].l.NodeNum = i;
   adjlist[k].r.NodeNum = j;
   adjlist1[k].l.NodeNum = i;
   adjlist1[k].r.NodeNum = j;

   adjlist[k].l.NodeInUse = YES;
   adjlist[k].r.NodeInUse = YES;
   adjlist1[k].l.NodeInUse = YES;
   adjlist1[k].r.NodeInUse = YES;

  histo[i]++; histo[j]++;
  }

 fclose(fd);

 for(i=1;i<=n;i++)
   {
    hhisto[histo[i]]++;
   } /* end for i */

 printf("\nHistogram of the number of nodes with i edges \n\n");
 for(i=1;i<=20;i++)
   {
    printf("%d --- \thhisto[%d]=%d \n",i,i,hhisto[i]);
    fflush(stdout);
   } /* end for i */

/* allocate memory for remining vectors */

  tableau = (Tableau *) calloc((n+1),sizeof(Tableau));
  if (tableau == NULL)
   {
    printf("can't allocate memory for tableau[] \n");
    exit(1);
   } /* end if */

  nodelist = (NodeList *) calloc((n+1),sizeof(NodeList));
  if (nodelist == NULL)
   {
    printf("can't allocate memory for nodelist[] \n");
    exit(1);
   } /* end if */

  SparseMatrix = (SparseMatrixSet *) calloc((n+1),sizeof(SparseMatrixSet));
  if (SparseMatrix == NULL)
   {
    printf("can't allocate memory for SparseMatrix[] \n");
    exit(1);
   } /* end if */

/* ititialize the SparseMatrix sets */
  for (i=1;i<=n;i++)
    {
     SparseMatrix[i].Row = NULL;
     name = (char *) calloc(10,sizeof(char));
     if (name == NULL)
      {
       printf("can't allocate memory for name[] in InputNetwork()\n");
       exit(1);
      } /* end if */
     name[0]  = 'R';
     name[1]  = 'o';
     name[2]  = 'w';
     name[3]  = '-';
     name[4]  = (char) (((i%100000)/10000) + 48);
     name[5]  = (char) (((i%10000)/1000) + 48);
     name[6]  = (char) (((i%1000)/100) + 48);
     name[7]  = (char) (((i%100)/10)  + 48);
     name[8]  = (char)  ((i%10)  + 48);
     name[9] = '\0';
     SparseMatrix[i].Row = InitSet(SparseMatrix[i].Row,name);
/*
PrintSet(SparseMatrix[i].Row);
*/
    } /* end for i */

/* to produce a legible X display reduce the image to less than 1k X 1k */
  ncomb = (n/1024)+1;
  ndisplay = ((n-1)/ncomb)+1;
printf("ncomb=%d -- ndisplay=%d \n",ncomb,ndisplay);
  temp = (short *) calloc(((ndisplay+1)*(ndisplay+1)),sizeof(short));
  if (temp == NULL)
   {
    printf("can't allocate memory for temp when allocating bitmap[][] \n");
    exit(1);
   } /* end if */
  bitmap = (short **) calloc((ndisplay+1),sizeof(short *));
  if (bitmap == NULL)
   {
    printf("can't allocate memory for bitmap[][] \n");
    exit(1);
   } /* end if */
  for (i=0; i<(ndisplay+1); i++)
    bitmap[i] = temp + i*(ndisplay+1);

} /* end InputNetwork() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ResetAdjList()
{
  int i;

 for (i=1;i<=nadjlist;i++)
  {
   adjlist[i].l.NodeNum = adjlist1[i].l.NodeNum;
   adjlist[i].r.NodeNum = adjlist1[i].r.NodeNum;

   adjlist[i].l.NodeInUse = YES;
   adjlist[i].r.NodeInUse = YES;
  } /* end for i */

} /* end ResetAdjList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void IdClusters()
{
 int i,j,ncluster,nnodelist,locmin,node,nextnode,nN2;
 int MinConNum;
 char *name;
 Set *N1i,*N1i_1,*Temp0;
 SetElement *pres;

/* initialize the sets N1i and N2 and set of sets N1 */
  N2 = NULL;
  N2 = InitSet(N2,"BorderEquations");
  N1 = NULL;
  N1 = InitSetOfSets(N1,"Sets of Clusters");
/* initialize the node list that identifies nodes that remain in the
   adjacency list and the corresponding adjacent cut set information */
  InitNodeList();
  nnodelist = n;
  ncluster = 0;

/* look for clusters until there are insufficient nodes to produce
   a cluster of minimum size */
  while (nnodelist > 0)
   {
/* initialize the tableau */
    InitTableau();
    ntab = 0;
    ncluster++;
/* initialize variables in preparation to search for a cluster */
    name = MakeClusterName(ncluster);
    N1i = NULL;
    N1i = InitSet(N1i,name);
/* identify the node with the minimum adjacent cut set */
    nextnode = StartNode();
/* add to the tableau as long as the sizeof the tableau is less
   than the maximum cluster size and there are nodes
   remaining in the node list */
    while ( (ntab < nmax) && (nnodelist > 0) )
     {
      if (nextnode <= 0)
       {
        printf("(nextnode <= 0) so this is a dead end \n");
        printf("ntab=%d  --  nnodelist=%d  --  nextnode=%d\n",ntab,nnodelist,nextnode);
        fflush(stdout);
        break;
       } /* end if nextnode < 0 */
      ntab++;
      nnodelist--;
      if (ntab == 1)
        N1i_1 = NULL;
      else
        N1i_1 = tableau[ntab-1].AdjacentSet;
      UpdateTableau(nextnode,N1i_1);
      node = nextnode;
/* identify whether or not the tableau can be expanded any further */
      if (tableau[ntab].ContourNum == 0) /* this is the dead-end problem */
       {
/*
        printf("tableau[%d].ContourNum = 0 \n",ntab); 
*/
        locmin = ntab;
        break;
       } /* end if tableau[ntab].NodeDegree == 0 */	
/* identify whether or not the tableau is sufficiently large as to look
   for a cluster break */
      if (ntab == nmin)
       {
        locmin = ntab;
        MinConNum = n;
       } /* end if */
      else if (ntab <= (3*nmax/4))
       {
        if (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum)
         {
          locmin = ntab;
         } /* end if */
       } /* end else if */
      else
       {
        if ( (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum) &&
             (tableau[ntab].ContourNum <= MinConNum) )
         {
          locmin = ntab;
          MinConNum = tableau[ntab].ContourNum;
         } /* end if */
       } /* end else */
      nextnode = SelectNextNodeInTableau();

      UpdateAdjacencyList(node);
     } /* end while ntab < nmax */
/* implement dynamic contour cutting by adding those nodes to N2 
   (the set of coupling equations) remaining after identifying a cluster */
    if (ntab <= nmin) locmin = ntab;
    if (locmin < ntab)
     {
/*
      printf("\n>\n>locmin=%d is less than ntab=%d - so update nnodelist \n",locmin,ntab);
*/
      nnodelist += (ntab-locmin);
      for (i=(locmin+1); i<=ntab; i++)
       {
        ResetAdjacencyList(tableau[i].NodeNum);
       } /* end for i */
     } /* end if */
    Temp0 = tableau[locmin].AdjacentSet;
/*
    PrintSet(Temp0);
*/
    nN2 = N2->n;
    UnionOfSets(N2,Temp0);
    nnodelist -= (N2->n - nN2);
/*
    PrintSet(N2);
*/

/* produce the set N1i by collecting all elements in the tableau into
   a set */
    for (i=1; i<=locmin; i++)
     {
      AddElementToSet(N1i,tableau[i].NodeNum);
     } /* end for i */

/* update the adjacency list to reflect removing both
   the cluster and the dynamic contour cut nodes */
    pres = Temp0->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */
    pres = N1i->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */

/* reset nodelist[].NodeNum and nodelist[].NodeInUse */
    InitNodeList();

/* add the set N1i to N1 - the set of cluster sets */
    AddSetToSetOfSets(N1,N1i);
/*
    PrintSet(N1i);
    PrintSet(N2);
*/
   } /* end while nnodelist > 0 */

/* add remaining nodes to the set N2 */
  for (i=1; i<=n; i++)
   {
    if (nodelist[i].NodeInUse == YES)
     {
      AddElementToSet(N2,i);
     } /* end if */
   } /* end for i */
/*
  PrintSetOfSets(N1);
  PrintSet(N2);
*/
} /* end IdClusters() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void SubClusterBlocks()
{
  int i,j;
  Set *N1i,*RN1i,*RN2;
  SetElement *element;
  SoSets *RN1,*TN1;
  SoSElement *soselN1,*soselRN1;

  TN1  = N1;

  N1  = NULL;
  N1  = InitSetOfSets(N1 ,TN1->Name);
  rN1 = NULL;
  rN1 = InitSetOfSets(rN1,TN1->Name);
/*
printf("SubClusterBlocks  --  set of clusters = |%s| \n",TN1->Name);
fflush(stdout);
*/
  soselN1 = TN1->First;
  while (soselN1 != NULL)
   {
    N1i = soselN1->set;
    AddSetToSetOfSets(N1,N1i);
/*
printf("SubClusterBlocks  --  N1i->Name=|%s| N1i->n=%d\n",N1i->Name,N1i->n);
fflush(stdout);
*/
    if ((N1i->n) > nmin)
     {
/* reset adjlist.x.NodeInUse to reflect only those nodes in the cluster */
      ResetNodesInN1i(N1i);
      InitNodeList();

/* initialize the set of sets RN1 and the set RN2 */
      RN2 = NULL;
      RN2 = InitSet(RN2,"ClusterBorderEquations");
      RN1 = NULL;
      RN1 = InitSetOfSets(RN1,N1i->Name);

/* generate the subclusters */
/*
printf("generate the subclusters for cluster = |%s| \n",N1i->Name);
fflush(stdout);
*/
      IdSubClusters(N1i,RN1,RN2);
/*
      PrintSet(N1i);
      PrintSetOfSets(RN1);
      PrintSet(RN2);
*/
/* reorder N1 to reflect the subclustering */
/* place the set RN2 into rN1 */
      AddSetToSetOfSets(rN1,RN2);
/* place the values in each subcluster into descending degree order */
      soselRN1 = RN1->First;
      while (soselRN1 != NULL)
       {
        RN1i = soselRN1->set;
/* place the set RN1i back into rN1i */
        AddSetToSetOfSets(rN1,RN1i);
        soselRN1 = soselRN1->Next;
       } /* end while */
     } /* end if */
    else
     {
/* place the set N1i into N1 */
/* place the values in N1i back into rN1 */
      AddSetToSetOfSets(rN1,N1i);
     } /* end else */

    soselN1 = soselN1->Next;
   } /* end while */

} /* end SubClusterBlocks() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ResetNodesInN1i(N1i)
  Set *N1i;
{
  int i;
  SetElement *element;

/* reset adjlist.x.NodeInUse to reflect only those nodes in the cluster */
/* set all nodes in edge conbinations to NO */
  for (i=1; i<=nadjlist; i++)
   {
    adjlist[i].l.NodeInUse = NO;
    adjlist[i].r.NodeInUse = NO;
   } /* end for i */
/* set all nodes in clusters to YES */
  element = N1i->First;
  while (element != NULL)
   {
    for (i=1; i<=nadjlist; i++)
     {
      if (adjlist[i].l.NodeNum == element->Element)
        adjlist[i].l.NodeInUse = YES;
      if (adjlist[i].r.NodeNum == element->Element)
        adjlist[i].r.NodeInUse = YES;
     } /* end for i */
    element = element->Next;
   } /* end while */
/* check to see that both nodes in an edge are in the cluster */
  for (i=1; i<=nadjlist; i++)
   {
    if ( (adjlist[i].l.NodeInUse !=YES) || (adjlist[i].r.NodeInUse !=YES) )
     {
      adjlist[i].l.NodeInUse = NO;
      adjlist[i].r.NodeInUse = NO;
     } /* end if */
   } /* end for i */
} /* end ResetNodesInN1i() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void IdSubClusters(N1i,RN1,RN2)
  Set *N1i,*RN2;
  SoSets *RN1;
{
 int i,j,ncluster,nnodelist,locmin,node,nextnode,nRN2;
 int MinConNum;
 char *name;
 Set *RN1i,*RN1i_1,*Temp0;
 SetElement *pres;

  nnodelist = N1i->n;
  ncluster = 0;

/* look for clusters */
  while (nnodelist > 0)
   {
/* initialize the tableau */
    InitTableau();
    ntab = 0;
    ncluster++;
/* initialize variables in preparation to search for a cluster */
    name = MakeSubClusterName(ncluster);
    RN1i = NULL;
    RN1i = InitSet(RN1i,name);
/* identify the node with the minimum adjacent cut set */
    nextnode = StartNode();
/* add to the tableau as long as the sizeof the tableau is less
   than the maximum cluster size and there are nodes
   remaining in the node list */
    while ( (ntab < nmin) && (nnodelist > 0) )
     {
      if (nextnode <= 0)
       {
        printf("(nextnode <= 0) so this is a dead end \n");
        printf("ntab=%d  --  nnodelist=%d  --  nextnode=%d\n",ntab,nnodelist,nextnode);
        fflush(stdout);
        break;
       } /* end if nextnode < 0 */
      ntab++;
      nnodelist--;
      if (ntab == 1)
        RN1i_1 = NULL;
      else
        RN1i_1 = tableau[ntab-1].AdjacentSet;
/*
printf("UpdateTableau  --  nextnode = %d\n",nextnode);
fflush(stdout);
*/
      UpdateTableau(nextnode,RN1i_1);
      node = nextnode;
/* identify whether or not the tableau can be expanded any further */
      if (tableau[ntab].ContourNum == 0) /* this is the dead-end problem */
       {
/*
        printf("tableau[%d].ContourNum = 0 \n",ntab);
*/
        locmin = ntab;
        break;
       } /* end if tableau[ntab].NodeDegree == 0 */
/* identify whether or not the tableau is sufficiently large as to look
   for a cluster break */
      if (ntab == 1) 
       {
        locmin = ntab;
        MinConNum = n;
       } /* end if */
      else if (ntab <= (nmin/2)) 
       {
        if (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum)
         {
          locmin = ntab;
         } /* end if */
       } /* end else if */
      else
       {
        if ( (tableau[ntab].ContourNum <= tableau[ntab-1].ContourNum) &&
             (tableau[ntab].ContourNum <= MinConNum) )
         {
          locmin = ntab;
          MinConNum = tableau[ntab].ContourNum;
         } /* end if */
       } /* end else */
      nextnode = SelectNextNodeInTableau();
/*
printf("nextnode = SelectNextNodeInTableau() = %d\n",nextnode);
fflush(stdout);
*/
      UpdateAdjacencyList(node);
     } /* end while ntab <= (nmax/8) */
/* implement dynamic contour cutting by adding those nodes to RN2
   (the set of coupling equations) remaining after identifying a cluster */
    if (ntab == 1) locmin = ntab;
    if (locmin < ntab)
     {
/*
      printf("\n>\n>locmin=%d is less than ntab=%d - so update nnodelist \n",locmin,ntab);
*/
      nnodelist += (ntab-locmin);
      for (i=(locmin+1); i<=ntab; i++)
       {
        ResetAdjacencyList(tableau[i].NodeNum);
       } /* end for i */
     } /* end if */
    Temp0 = tableau[locmin].AdjacentSet;
/*
    PrintSet(Temp0);
*/
    nRN2 = RN2->n;
    UnionOfSets(RN2,Temp0);
    nnodelist -= (RN2->n - nRN2);
/*
    PrintSet(RN2);
*/

/* produce the set RN1i by collecting all elements in the tableau into
   a set */
    for (i=1; i<=locmin; i++)
     {
      AddElementToSet(RN1i,tableau[i].NodeNum);
     } /* end for i */

/* update the adjacency list to reflect removing both
   the cluster and the dynamic contour cut nodes */
    pres = Temp0->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */
    pres = RN1i->First;
    while (pres != NULL)
     {
      UpdateAdjacencyList(pres->Element);
      pres = pres->Next;
     } /* end while pres != NULL */

/* reset nodelist[].NodeNum and nodelist[].NodeInUse */
    InitNodeList();

/* add the set RN1i to RN1 - the set of cluster sets */
    AddSetToSetOfSets(RN1,RN1i);
   } /* end while nnodelist > 0 */

/* add remaining nodes to the set RN2 */
  for (i=1; i<=n; i++)
   {
    if (nodelist[i].NodeInUse == YES)
     {
      AddElementToSet(RN2,i);
     } /* end if */
   } /* end for i */
/*
  PrintSetOfSets(RN1);
  PrintSet(RN2);
*/
} /* end IdSubClusters() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeClusterName(ncluster)
  int  ncluster;
{
  char *name;
  char *string = "Cluster";
  int i;

  name = (char *) calloc(12,sizeof(char));
  if (name == NULL)
   {
    printf("can't allocate memory for name[] in MakeClusterName()\n");
    exit(1);
   } /* end if */

  for (i=0; i<7; i++)
   name[i] = string[i];

  name[7]  = '-';
  name[8]  = (char) (((ncluster%1000)/100) + 48);
  name[9]  = (char) (((ncluster%100)/10)  + 48);
  name[10] = (char) ((ncluster%10)  + 48);
  name[11] = '\0';

} /* end MakeClusterName(); */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
char *MakeSubClusterName(ncluster)
  int  ncluster;
{
  char *name;
  char *string = "SubCluster";
  int i;

  name = (char *) calloc(15,sizeof(char));
  if (name == NULL)
   {
    printf("can't allocate memory for name[] in MakeSubClusterName()\n");
    exit(1);
   } /* end if */

  for (i=0; i<10; i++)
   name[i] = string[i];

  name[10]  = '-';
  name[11]  = (char) (((ncluster%1000)/100) + 48);
  name[12]  = (char) (((ncluster%100)/10)  + 48);
  name[13] = (char) ((ncluster%10)  + 48);
  name[14] = '\0';

} /* end MakeClusterName(); */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InitNodeList()
{
  int i;

/* initialize nodelist[].NodeNum & nodelist[].NodeInUse */
  for (i=1;i<=n;i++)
   {
    nodelist[i].NodeNum = 0;
    nodelist[i].NodeInUse = NO;
   } /* end for i */

/* calculate the number of connections at each node */

  for (i=1;i<=nadjlist;i++)
   {
    if (adjlist[i].l.NodeInUse == YES)
     {
      (nodelist[ adjlist[i].l.NodeNum ].NodeNum)++;
      nodelist[ adjlist[i].l.NodeNum ].NodeInUse = YES;
     } /* end if */
    if (adjlist[i].r.NodeInUse == YES)
     {
      (nodelist[ adjlist[i].r.NodeNum ].NodeNum)++;
      nodelist[ adjlist[i].r.NodeNum ].NodeInUse = YES;
     } /* end if */
   } /* end for i */

/*
  for (i=1;i<=n;i++)
   {
printf("nodelist[%d].NodeNum=%d  --  nodelist[%d].NodeInUse=%d\n",i,nodelist[i].NodeNum,i,nodelist[i].NodeInUse);
fflush(stdout);
   } /* end for i */
} /* end for InitNodeList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void InitTableau()
{
  int i;

  for (i=1; i<=n; i++)
   {
    tableau[i].NodeNum = 0;
    tableau[i].ContourNum = 0;
    if (tableau[i].AdjacentSet != NULL)
      DeleteSet(tableau[i].AdjacentSet);
   } /* end for i */
} /* end InitTableau() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
int StartNode()
{
 int i,MinDeg,MinDegNode;

/* find the minimum degree node */

  MinDeg = n;
  MinDegNode = 0;

  for (i=1;i<=n;i++)
   {
    if ( (nodelist[i].NodeNum < MinDeg) && (nodelist[i].NodeInUse == YES) )
     {
      MinDeg = nodelist[i].NodeNum;
      MinDegNode = i;
     } /* end if */
   } /* end for i */

if (MinDegNode == 0)
 {
  printf("In StartNode() -- MinDegNode = 0 \n");
  fflush(stdout);
 } /* end if */

 return MinDegNode;

} /* end StartNode() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void UpdateTableau(nextnode,N1i_1)
  int nextnode;
  Set *N1i_1;
{
  int i;
  Set *adjset;
  SetElement *pres;

  tableau[ntab].NodeNum = nextnode;

/* generate the set of adjacent nodes to this node */
/* first - find all node adjacent to this node */
  adjset = NULL;
  adjset = InitSet(adjset,"Adjacent Nodes");
  for (i=1; i<=nadjlist; i++)
   {
/*
printf("UpdateTableau() -- i=%d \n",i);
fflush(stdout);
*/
    if( (adjlist[i].l.NodeNum == nextnode) && 
        (adjlist[i].l.NodeInUse == YES)    &&
        (adjlist[i].r.NodeInUse == YES) )
     {
/*
printf("AddElementToSet(adjset,%d) \n",adjlist[i].r.NodeNum);
fflush(stdout);
*/
      AddElementToSet(adjset,adjlist[i].r.NodeNum);
     } /* end if */
    if( (adjlist[i].r.NodeNum == nextnode) && 
        (adjlist[i].r.NodeInUse == YES)    &&
        (adjlist[i].l.NodeInUse == YES) )
     {
/*
printf("AddElementToSet(adjset,%d) \n",adjlist[i].l.NodeNum);
fflush(stdout);
*/
      AddElementToSet(adjset,adjlist[i].l.NodeNum);
     } /* end if */
   } /* end for i */

  if (ntab > 1)
   {
/*
printf("print out the adjacent set from the previous tableau entry \n");
PrintSet(N1i_1);
printf("print out the partial adjacent set from this tableau entry \n");
PrintSet(adjset);
*/
/* second - take the union of the set of adjacent nodes with the set of 
   adjacent nodes at the previous node in the adjacent set */
    UnionOfSets(adjset,N1i_1);
/* third - remove this node from the adjacent set */
    DeleteElementFromSet(adjset,nextnode);
/*
printf("print out the updated adjacent set from this tableau entry \n");
fflush(stdout);
PrintSet(adjset);
*/
   } /* end if (ntab > 1) */

  tableau[ntab].ContourNum = adjset->n;

/* print the header */
/*
  printf("%d - NodeNum=%d - AdjSet={",
          ntab,tableau[ntab].NodeNum);
  pres = adjset->First;
  while (pres != NULL)
   {
    printf("(%d,%d)",pres->Element,nodelist[pres->Element].NodeNum);
    pres = pres->Next;
   } /* end while */
/*
  printf("} - ContourNum=%d \n",tableau[ntab].ContourNum);
*/
  tableau[ntab].AdjacentSet = adjset;
  
} /* end UpdateTableau() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
int SelectNextNodeInTableau()
{
  int i,node,MinDeg,MinDegNode;
  SetElement *pres;
  Set *set;

/* check each element in the set to determine which element has the
   fewest connecting nodes given that this node hasn`t already been
   chosen.  If there are ties chose the first value encountered */

  MinDeg = n;
  MinDegNode = 0;

  set = tableau[ntab].AdjacentSet;
  pres = set->First;
  while (pres != NULL)
   {
    node = pres->Element;
    if ( (nodelist[node].NodeNum < MinDeg) && (nodelist[node].NodeInUse == YES) )
     {
      MinDeg = nodelist[node].NodeNum;
      MinDegNode = node;
     } /* end if */
    pres = pres->Next;
   } /* end while pres != NULL */

  return MinDegNode;

} /* end SelectNextNodeInTableau() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ResetAdjacencyList(node)
  int node;
{
  int i;

/* mark this node as again accessible for further consideration */
  for (i=1; i<=nadjlist; i++)
   {
    if (node == adjlist[i].l.NodeNum)
      adjlist[i].l.NodeInUse = YES;
    if (node == adjlist[i].r.NodeNum)
      adjlist[i].r.NodeInUse = YES;
   } /* end for i */

} /* end ResetAdjacencyList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void UpdateAdjacencyList(node)
  int node;
{
  int i;

/* mark this node as not accessible for further consideration */
  for (i=1; i<=nadjlist; i++)
   {
    if (node == adjlist[i].l.NodeNum)
      adjlist[i].l.NodeInUse = NO;
    if (node == adjlist[i].r.NodeNum)
      adjlist[i].r.NodeInUse = NO;
   } /* end for i */

} /* end UpdateAdjacencyList() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ReorderNetworkClusters(TN1,namekey)
  SoSets *TN1;
  char *namekey;
{
  int i,j,k,here,iproc,jproc,icolor;
  int MinDeg,MinDegLoc,nN1i,nN2,node;
  int nlastblockstart,nlastblockend;
  Set *N1i;
  SetElement *element,*pres;
  SoSElement *soselement;
  char filename[80];

printf("ReorderNetworkClusters --- TN1->n=|%d| TN1->Name=|%s| \n",
                                   TN1->n,TN1->Name);
printf("\t\tn=%d - nadjlist=%d \n",n,nadjlist);
fflush(stdout);

/* Check the integrity of the clustered data */
  for (j=1;j<=n;j++)
   {
    here = NO;
    soselement = TN1->First;
    while (soselement != NULL)
     {
      N1i = soselement->set;
      if (ElementOfSet(N1i,j) == YES)
       {
        here = YES;
        break;
       } /* end if */
      soselement = soselement->Next;
     } /* end while soselement != NULL */
    if (here == NO)
     {
      if (ElementOfSet(N2,j) == NO)
        printf("** -- row %d is neither in a cluster nor in the border \n",j);
     } /* end if (here == NO) */
   } /* end for j */

  if (ReorderLst != NULL) free(ReorderLst);
  ReorderLst = (int *) calloc((n+1),sizeof(int));
  if (ReorderLst == NULL)
   {
    printf("can't allocate memeory for ReorderLst[] \n");
    exit(1);
   } /* end if */

/* order into clusters */
  i=0;
  soselement = TN1->First;
  while (soselement != NULL)
   {
    N1i = soselement->set;
    nN1i = N1i->n;
/*
printf("Cluster set -- |%s| nN1i = %d \n",N1i->Name,nN1i);
fflush(stdout);
*/
/*
PrintSet(N1i);
fflush(stdout);
*/
    element = N1i->First;
    while (element != NULL)
     {
      node = element->Element;
/*
printf("Within loop -------node = %d\n",node);
fflush(stdout);
*/
      ReorderLst[node] = ++i;
      element = element->Next;
/*
printf("Within loop -- j = %d --- element = %d\n",j,element);
fflush(stdout);
*/
     } /* end while element != NULL */
    soselement = soselement->Next;
   } /* end while soselement != NULL */

/* initially order the last diagonal block */

printf("initially order the last diagonal block \n");
printf("Border set -- |%s|  - (N2->n)=%d \n",N2->Name,N2->n);

/*
PrintSet(N2);
*/
  element = N2->First;
  while (element != NULL)
   {
    node = element->Element;
    ReorderLst[node] = ++i;
    element = element->Next;
   } /* end while element != NULL */

/* exchange the node numbers in the adjacency list to reorder the matrix */
/*
printf("reorder the matrix \n");
fflush(stdout);
*/
  for (i=1;i<=nadjlist;i++)
   {
    adjlist[i].l.NodeNum = ReorderLst[adjlist[i].l.NodeNum];
    adjlist[i].r.NodeNum = ReorderLst[adjlist[i].r.NodeNum];
   } /* end for i */ 

printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap("none"); 

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the graph */

/* generate adjlistLB[] - first pass is to identify the number of nodes */
    nlastblockstart = n - (N2->n) + 1;
    nlastblockend   = n;
    printf("ln=%d - nlastblockstart=%d nlastblockend=%d \n",(N2->n),nlastblockstart,nlastblockend);
    nadjlistLB = 0;
    for (i=1;i<=nadjlist;i++)
      {
        if ( (adjlist[i].l.NodeNum >= nlastblockstart) && 
             (adjlist[i].r.NodeNum >= nlastblockstart))
	  {
           nadjlistLB++;
           printf("\t%d \t%d %d \n",nadjlistLB,adjlist[i].l.NodeNum,adjlist[i].r.NodeNum);
	  } /* end if */
      } /* end for i */

/* allocate memory for the adjacency list for the last block */
   if(adjlistLB != NULL) free(adjlistLB);
   adjlistLB = (AdjListLB *) calloc((nadjlistLB+1),sizeof(AdjListLB));
   if (adjlistLB == NULL)
    {
     printf("can't allocate memory for adjlistLB[] \n");
     exit(1);
    } /* end if */

/* load data into adjlistLB */

printf("load data into adjlistLB \n");
fflush(stdout);

    nadjlistLB = 0;
    for (i=1;i<=nadjlist;i++)
      {
        if ( (adjlist[i].l.NodeNum >= nlastblockstart) && 
             (adjlist[i].r.NodeNum >= nlastblockstart))
	  {
           nadjlistLB++;
           adjlistLB[nadjlistLB].l.NodeNum  =  adjlist[i].l.NodeNum;
           adjlistLB[nadjlistLB].l.NodeNumR = (adjlist[i].l.NodeNum - nlastblockstart + 1);
           adjlistLB[nadjlistLB].r.NodeNum  =  adjlist[i].r.NodeNum;
           adjlistLB[nadjlistLB].r.NodeNumR = (adjlist[i].r.NodeNum - nlastblockstart + 1);
	  } /* end if */
      } /* end for i */

/* allocate memory for nodeLB[] */
 
printf("allocate memory for nodeLB[] \n");
fflush(stdout);

   if (nodeLB != NULL) free(nodeLB);
   nodeLB = (NodeLB *) calloc(((N2->n)+1),sizeof(NodeLB));
   if (nodeLB == NULL)
    {
     printf("can't allocate memory for nodeLB[] \n");
     exit(1);
    } /* end if */

/* allocate memory for nodeLBT[] */
 
printf("allocate memory for nodeLBT[] \n");
fflush(stdout);

   if (nodeLBT != NULL) free(nodeLBT);
   nodeLBT = (NodeLB *) calloc(((N2->n)+1),sizeof(NodeLB));
   if (nodeLBT == NULL)
    {
     printf("can't allocate memory for nodeLBT[] \n");
     exit(1);
    } /* end if */

printf("display the last diagonal block - namekey=|%s| \n",namekey);
fflush(stdout);

  GenerateBitmapLB(MakeOutputFilenameLB(namekey,filename,BEFORE_COLORING),BEFORE_COLORING);
  PrintLB("Last Block Before Coloring");

/* color the last diagonal block */

printf("color the last diagonal block \n");
fflush(stdout);

  ColorLB();

/* allocate memory for ulbbufsize[] */
  
printf("allocate memory for ulbbufsize[] \n");
fflush(stdout);

  if (ulbbufsizetmp != NULL) free(ulbbufsizetmp);
   ulbbufsizetmp = (int *) calloc(((nprocs+1)*(nprocs+1)),sizeof(int));
   if (ulbbufsizetmp == NULL)
    {
     printf("can't allocate memory for ulbbufsizetmp[] \n");
     exit(1);
    } /* end if */
   if (ulbbufsize != NULL) free(ulbbufsize);
   ulbbufsize = (int **) calloc((nprocs+1),sizeof(int *));
   if (ulbbufsize == NULL)
    {
     printf("can't allocate memory for ulbbufsize[][] \n");
     exit(1);
    } /* end if */
   for (iproc=0;iproc<=nprocs;iproc++) 
    ulbbufsize[iproc] = ulbbufsizetmp + (iproc*(nprocs+1));
   for (iproc=0;iproc<=nprocs;iproc++) 
    for (jproc=0;jproc<=nprocs;jproc++)
      ulbbufsize[iproc][jproc] = 0;

/* allocate memory for commlb[][] */
 
printf("allocate memory for commlb[][] \n");
fflush(stdout);

   if (commlb != NULL) free(commlb);
   commlb = (CommLB **) calloc((ncolors+1),sizeof(CommLB *));
   if (commlb == NULL)
    {
     printf("can't allocate memory for commlb[][] \n");
     exit(1);
    } /* end if */
   if (commlbtmp != NULL) free(commlbtmp);
   commlbtmp = (CommLB *) calloc(((nprocs+1)*(ncolors+1)),sizeof(CommLB));
   if (commlbtmp == NULL)
    {
     printf("can't allocate memory for commlbtmp[] \n");
     exit(1);
    } /* end if */
   for (icolor=0;icolor<=ncolors;icolor++) commlb[icolor] = commlbtmp + icolor*(nprocs+1);
   for (icolor=1;icolor<=ncolors;icolor++)
    for (iproc=0;iproc<=nprocs;iproc++)
      commlb[icolor][iproc].bufsize = 0;

/* order the last diagonal block and border to account for load balancing */
/* do not change the block diagonal partitions */
  for (i=1;i<=n;i++)
      ReorderLst[i] = i;

/* divide colors equally between processors */
  iproc = 0;
  for (icolor=1;icolor<=ncolors;icolor++)
   {
    for (j=1;j<=(N2->n);j++)
     {
      if (nodeLB[j].Color == icolor)
	{
/*
printf("divide colors --- icolor=%d - j=%d - iproc=%d \n",icolor,j,iproc);
fflush(stdout);
*/
         nodeLB[j].ProcNum = (iproc+1);
         (commlb[icolor][iproc+1].bufsize)++;
         iproc = (iproc+1)%nprocs;
	} /* end if */
     } /* end for i */
   } /* end for icolor */

/* transfer the data from nodeLB to nodeLBT */
  for (j=1;j<=(N2->n);j++)
    {
     nodeLBT[j].NodeNum           = nodeLB[j].NodeNum;
     nodeLBT[j].NodeNumR          = nodeLB[j].NodeNumR;
     nodeLBT[j].Color             = nodeLB[j].Color;
     nodeLBT[j].ProcNum           = nodeLB[j].ProcNum;
     nodeLBT[j].Neighbors         = nodeLB[j].Neighbors;
     nodeLBT[j].ColoredNeighbors  = nodeLB[j].ColoredNeighbors;
     nodeLBT[j].ColorsOfNeighbors = nodeLB[j].ColorsOfNeighbors;
    } /* end for j */

/* reorder the borders and last diagonal block */

printf("reorder the last diagonal block and border \n");
fflush(stdout);

  j = 0;
  i = (nlastblockstart-1);
  for (icolor=1;icolor<=ncolors;icolor++)
   {
    for (iproc=1;iproc<=nprocs;iproc++)
      {
        for (node=1;node<=(N2->n);node++)
          {
           if ( (nodeLBT[node].ProcNum == iproc) && (nodeLBT[node].Color == icolor) )
	    {
/*
printf("\treorder --- j=%d - iproc=%d - node=%d - nodeLBT[%d].ProcNum=%d - icolor=%d - i=%d - loc=%d \n",
         ++j,iproc,node,node,nodeLBT[node].ProcNum,nodeLBT[node].Color,(i+1),(node+nlastblockstart-1));
fflush(stdout);
*/
             ReorderLst[node+nlastblockstart-1] = ++i;
             nodeLB[j].NodeNum           = nodeLBT[node].NodeNum;
             nodeLB[j].NodeNumR          = nodeLBT[node].NodeNumR;
             nodeLB[j].Color             = nodeLBT[node].Color;
             nodeLB[j].ProcNum           = nodeLBT[node].ProcNum;
             nodeLB[j].Neighbors         = nodeLBT[node].Neighbors;
             nodeLB[j].ColoredNeighbors  = nodeLBT[node].ColoredNeighbors;
             nodeLB[j].ColorsOfNeighbors = nodeLBT[node].ColorsOfNeighbors;
	    } /* end if */
          } /* end for node */
      } /* end for iproc */
   } /* end for icolor */

/* update commlb[][].starti */
  i = 1;
  for (icolor=1;icolor<=ncolors;icolor++)
    for (iproc=1;iproc<=nprocs;iproc++)
     {
       commlb[icolor][iproc].starti = i;
       i += commlb[icolor][iproc].bufsize;
     } /* end for iproc & icolor */

/* exchange the node numbers in the adjacency list to reorder the matrix */

printf("reorder the matrix considering the coloring in the last diagonal block \n");
fflush(stdout);

  for (i=1;i<=nadjlist;i++)
   {
    adjlist[i].l.NodeNum = ReorderLst[adjlist[i].l.NodeNum];
    adjlist[i].r.NodeNum = ReorderLst[adjlist[i].r.NodeNum];
   } /* end for i */ 

printf("display the last diagonal block after coloring and reordering \n");
fflush(stdout);

  GenerateBitmapLB(MakeOutputFilenameLB(namekey,filename,AFTER_COLORING),AFTER_COLORING);
  PrintLB("Last Block After Coloring");

/* load the data into the SparseMatrix sets */

printf("load the data into the SparseMatrix sets --- n=%d \n",n);
fflush(stdout);

  for (i=1;i<=n;i++)
   {
    AddElementToSet(SparseMatrix[i].Row,i);
   } /* end for i */
  for (i=1;i<=nadjlist;i++)
   {
/*
printf("i=%d - adjlist[%d].l.NodeNum=%d - adjlist[%d].r.NodeNum=%d \n",
           i,i,adjlist[i].l.NodeNum,i,adjlist[i].r.NodeNum);
fflush(stdout);
*/
    AddElementToSet(SparseMatrix[adjlist[i].l.NodeNum].Row,adjlist[i].r.NodeNum);
    AddElementToSet(SparseMatrix[adjlist[i].r.NodeNum].Row,adjlist[i].l.NodeNum);
   } /* end for i */

printf("load the data into the SparseMatrix sets --- n=%d \n",n);
fflush(stdout);

} /* end ReorderNetworkClusters() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void NonZeroDensityInLastBlock()
{
  int i,j,lbz,lbnonz;
  SetElement *pres;

  lbz = lbnonz = 0;
  for (i=1; i<=nadjlist; i++)
   {
    if ( (adjlist[i].l.NodeNum > (n-(N2->n)) ) && (adjlist[i].r.NodeNum > (n-(N2->n)) ) )
      lbnonz++;
   } /* end for i */

  lbz = ((N2->n) * ((N2->n)-1) / 2) - lbnonz;

  printf("The number of equations in the borders/last block = %d\n",N2->n);
  printf("The number of nonzeros in the last block          = %d\n",lbnonz);
  printf("The number of zeros in the last block             = %d\n",lbz);
  printf("The percentage of nonzeros                        = %f\n",
           ( (double) (lbnonz * 100) / (double) ( (((N2->n)-1)*(N2->n) / 2) ) ));
    
} /* NonZeroDensityInLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void GenerateBitmap(filename)
  char *filename;
{
 int i,j,k,ipixel;
 int idisplay,jdisplay;
 int PointerL;
 int Nonzero;

 FILE *fd;

  ntot=0;
/*
printf("\nIn GenerateBitmap() - n=%d \n",n);
*/

/* initialize the display bitmap to all white */
  for (i=1;i<=ndisplay;i++)
   for (j=1;j<=ndisplay;j++)
    bitmap[j][i] = ZERO;

/* enter the diagonal */
 for (k=1; k<=n; k++) /* for the diagonal */
  {
   idisplay = ((k-1)/ncomb)+1;
   jdisplay = ((k-1)/ncomb)+1;
   ntot++;
   bitmap[idisplay][jdisplay]=NONZERO;
  } /* end for k */

 for (k=1; k<=nadjlist; k++) /* for all edges */
  {
   i = adjlist[k].l.NodeNum;
   j = adjlist[k].r.NodeNum;
   idisplay = ((i-1)/ncomb)+1;
   jdisplay = ((j-1)/ncomb)+1;
   ntot += 2;
   bitmap[jdisplay][idisplay]=NONZERO;
   bitmap[idisplay][jdisplay]=NONZERO;
 } /* end for k */

printf("\nIn GenerateBitmap() - filename = |%s| - ntot=%d \n",filename,ntot);
fflush(stdout);

  if (strcmp(filename,"none") != 0)
   {
    fd = fopen(filename, "w");

printf("Open |%s| for output of the Bitmap\n",filename);
fflush(stdout);

    for (i=1;i<=ndisplay;i++)
     for (j=1;j<=ndisplay;j++)
      {
       ipixel = bitmap[j][i];
       if( (ipixel < 0) || (ipixel > 255) )
        {
printf("error in writting bitmap[%d][%d]=|%d| to %s\n",j,i,ipixel);
fflush(stdout);
         ipixel = 0;
        } /* end if */
       fputc(ipixel,fd);
      }

    fclose(fd);

printf("Close |%s| after output of the Bitmap\n",filename);
fflush(stdout);
  } /* end if filename != none */

} /* end GenerateBitmap() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void GenerateBitmapLB(filename,iselect)
  char *filename;
  int iselect;
{
 int i,j,k,ipixel,icolor,nlastblockstart;
 int ncombLB,ndisplayLB;
 int idisplay,jdisplay;
 int PointerL;
 int Nonzero;
 int Color[11];

 FILE *fd;

  ntot=0;
  Color[0]  = ZERO;
  Color[1]  = COLOR_1;
  Color[2]  = COLOR_2;
  Color[3]  = COLOR_3;
  Color[4]  = COLOR_4;
  Color[5]  = COLOR_5;
  Color[6]  = COLOR_6;
  Color[7]  = COLOR_7;
  Color[8]  = COLOR_8;
  Color[9]  = COLOR_9;
  Color[10] = COLOR_10;
  nlastblockstart = n - (N2->n) + 1;


printf("\nIn GenerateBitmapLB() - n=%d \n",n);


/* calculate the sizing factor */
  ncombLB = ((N2->n)/1024)+1;
  ndisplayLB = (((N2->n)-1)/ncombLB)+1;
printf("ncombLB=%d -- ndisplayLB=%d \n",ncombLB,ndisplayLB);

printf("initialize the display bitmap to all white  \n");
fflush(stdout);

/* initialize the display bitmap to all white */
  for (i=1;i<=ndisplayLB;i++)
   for (j=1;j<=ndisplayLB;j++)
    bitmap[j][i] = ZERO;

printf("enter the diagonal - (N2->n)=%d\n",(N2->n));
fflush(stdout);

/* enter the diagonal */
 for (k=1; k<=(N2->n); k++) 
  {
/*
printf("k=%d -", k);
fflush(stdout);
printf(" iselect=%d -",iselect);
fflush(stdout);
printf(" nodeLB[%d].Color=%d -",k,nodeLB[k].Color);
fflush(stdout);
printf(" nodeLB[%d].ProcNum=%d -",k,nodeLB[k].ProcNum);
fflush(stdout);
printf("- Color[nodeLB[%d].Color]=%d \n", k,Color[nodeLB[k].Color]);
fflush(stdout);
*/
   idisplay = ((k-1)/ncombLB)+1;
   jdisplay = ((k-1)/ncombLB)+1;
   ntot++;
/*
printf("k=%d - iselect=%d - nodeLB[%d].Color=%d - Color[nodeLB[%d].Color]=%d \n",
         k,iselect,k,nodeLB[k].Color,k,Color[nodeLB[k].Color]);
fflush(stdout);
*/
   if (iselect == BEFORE_COLORING)
     bitmap[idisplay][jdisplay] = NONZERO;
   else 
     bitmap[idisplay][jdisplay] = Color[nodeLB[k].Color];
  } /* end for k */

printf("enter the rest of the matrix - (N2->n)=%d\n",(N2->n));
fflush(stdout);

 for (k=1; k<=nadjlist; k++) /* for all edges */
  {
   i = adjlist[k].l.NodeNum;
   j = adjlist[k].r.NodeNum;
   if ( (i >= nlastblockstart) && (j >= nlastblockstart) )
     {
      idisplay = ((i-nlastblockstart)/ncombLB)+1;
      jdisplay = ((j-nlastblockstart)/ncombLB)+1;
      ntot += 2;
/*
printf("k=%d - i=%d - j=%d - idisplay=%d - jdisplay=%d\n",
         k,i,j,idisplay,jdisplay);
fflush(stdout);
*/
      if (iselect == BEFORE_COLORING)
	{
         bitmap[jdisplay][idisplay] = NONZERO;
         bitmap[idisplay][jdisplay] = NONZERO;
	} /* end if */
      else 
	{
/*
printf("\t Color[nodeLB[%d].Color]=%d - Color[nodeLB[%d].Color]=%d \n",
         jdisplay,Color[nodeLB[jdisplay].Color],idisplay,Color[nodeLB[idisplay].Color]);
fflush(stdout);
*/
          bitmap[jdisplay][idisplay] = Color[nodeLB[jdisplay].Color];
          bitmap[idisplay][jdisplay] = Color[nodeLB[idisplay].Color];
	} /* end else */
     } /* end if */
 } /* end for k */

printf("\nIn GenerateBitmapLB() - filename = |%s| ntot=%d \n",filename,ntot);
fflush(stdout);

  if (strcmp(filename,"none") != 0)
   {
    fd = fopen(filename, "w");

printf("Open |%s| for output of the Bitmap - ndisplayLB=%d \n",filename,ndisplayLB);
fflush(stdout);

    for (i=1;i<=ndisplayLB;i++)
     for (j=1;j<=ndisplayLB;j++)
      {
       ipixel = bitmap[j][i];
       if( (ipixel < 0) || (ipixel > 255) )
        {
printf("error in writting bitmap[%d][%d]=|%d| to %s\n",j,i,ipixel);
fflush(stdout);
         ipixel = 0;
        } /* end if */
       fputc(ipixel,fd);
      }

    fclose(fd);

printf("Close |%s| after output of the Bitmap\n",filename);
fflush(stdout);
  } /* end if filename != none */

} /* end GenerateBitmapLB() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void PrintL(title)
 char *title;
{
 int i,j;
 int PointerL;
 int Nonzero;

 printf("\nSummary statistics for matrix structure \n");
 printf("\tNumber of nonzeros in L                                    = %d \n",(n+nadjlist));
 printf("\tNumber of places in the entire matrix                      = %d \n",(n*n));
 printf("\tNumber of nonzeros in the strictly lower triangular matrix = %d \n",(nadjlist));
 printf("\tNumber of places in the strictly lower triangular matrix   = %d \n",((n-1)*n/2));
 printf("\tPercentage of non-zeros in the matrix                      = %f \n",
               ((double) (nadjlist * 100)/ (double) ((n-1)*n/2)));
 fflush(stdout);

 if (xdisplay == YES)
   {
    printf("\n\n start X-Window display %s\n\n",title);
    Ximage(ndisplay,ndisplay,title);
    printf("\n\n end X-Window display %s\n\n",title);
    fflush(stdout);
   } /* end if */

} /* end PrintL() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void PrintLB(title)
 char *title;
{
 if (xdisplay == YES)
   {
    printf("\n\n start X-Window display %s\n\n",title);
    Ximage((N2->n),(N2->n),title);
    printf("\n\n end X-Window display %s\n\n",title);
    fflush(stdout);
   } /* end if */

} /* end PrintLB() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void CountOpsBDB()
{
  int i,j,k,jk,ji,ik,blockstart,blockend;
  int nOpsTotalGS,nOpsBlockGS,nOpsBorderGS;
  int borderstart,borderend;
  SoSElement *soselement;
  SetElement *elementRowk,*elementRowi;
  Set *N1i;

/* initialize the values of block/border edges */
  blockstart  = blockend = 0;
  borderstart = (n - N2->n) + 1;
  borderend   = n;
  nOpsLastBlockGS = 0;
  nOpsTotalGS = nOpsBlockGS = nOpsBorderGS = 0;

/* allocate memory for the vector of independent blocks */
  i = (n - N2->n);
  if (IndBlk != NULL) free(IndBlk);
  IndBlk = (IndBlocks *) calloc((i+1),sizeof(IndBlocks));
  if (IndBlk == NULL)
   {
    printf("can't allocate memory for IndBlk[] \n");
    exit(1);
   } /* end if */

/* set up the vector IndBlk[] */
  j = 0;
  soselement = N1->First;
  while (soselement != NULL)
   {
    N1i = soselement->set;
    blockstart = blockend + 1;
    blockend += N1i->n;
    if ( N1i->n <= 0)
      {
       printf("For cluster %s - N1i->n=%d \n",N1i->Name,N1i->n);
       PrintSet(N1i);
       fflush(stdout);
      } /* end if */
/*
printf("for cluster %s  --  blockstart=%d  --  blockend=%d \n",N1i->Name,blockstart,blockend);
fflush(stdout);
*/
    for (k=blockstart; k<=blockend; k++) /* along diagonal */
     {
      j++;
      IndBlk[j].N1i = N1i;
      IndBlk[j].blockstart = blockstart;
      IndBlk[j].blockend   = blockend;
     } /* end for k */

    soselement = soselement->Next;
   } /* end while soselement != NULL */

  if (i != j)
    {
     printf("(i & j should be equal!) j=%d  --  i=%d \n",j,i);
     fflush(stdout);
    } /* end if */

/*        k  is the index for the diagonal
          i  is an element of column k
	  jk is an element of row k
	  ji is an element of row i
*/
/* calculate the number of operations for Gauss-Seidel */
  for (k=1; k<=n; k++) /* along diagonal */
  {
/*
printf("Calc ops in GS --- k=%d -  \n",k);
fflush(stdout);
*/
/* these data structures are for the rows in the strictly 
   lower diagonal portion of the matrix */
   elementRowk = SparseMatrix[k].Row->First;
   while (elementRowk != NULL)
     {
      jk = elementRowk->Element;
/*
printf("\t\tjk=%d -  \n",jk);
fflush(stdout);
*/
/* calculate the number of multiply/adds in this row of a block */
      if (k < borderstart)
	{
         (IndBlk[k].N1i->nOpsBlockGS)++;
         (IndBlk[k].N1i->nOpsTotalGS)++;
         nOpsBlockGS++;
         nOpsTotalGS++;
	} /* end if k is in independent blocks */
      else
/* calculate the number of multiply/adds in this row of a border */
	{
         if (jk < borderstart)
	   {
            (IndBlk[jk].N1i->nOpsBorderGS)++;
            (IndBlk[jk].N1i->nOpsTotalGS)++;
            nOpsBorderGS++;
            nOpsTotalGS++;
	   } /* end if jk < borderstart */
         else
/* calculate the number of multiply/adds in this row of the last block */
	   {
            nOpsTotalGS++;
            nOpsLastBlockGS++;
	   } /* end if jk >= borderstart */
	} /* end if k is in the border */
      elementRowk = elementRowk->Next;
     } /* end while elementRowk != NULL */
  } /* end for k */

  soselement = N1->First;
  while (soselement != NULL)
   {
    N1i = soselement->set;
    N1i->nOpsTotal = N1i->nOpsTotalGS;
    soselement = soselement->Next;
   } /* end while soselement !=NULL */

printf("\nOperations Counts for Gauss-Seidel \n");
printf("\tGS - nOpsBlockGS=%d \n",nOpsBlockGS);
printf("\tGS - nOpsBorderGS=%d \n",nOpsBorderGS);
printf("\tGS - nOpsTotalGS=%d \n",nOpsTotalGS);

printf("\n\t(LastBlock - nOpsLastBlockGS=%d \n",nOpsLastBlockGS);

printf("\nTotal operations Count=%d \n",nOpsTotalGS);
fflush(stdout);

  free(IndBlk);

} /* end CountOpsBBD() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void PigeonHoleLoadBalance(tN1,FileNameA,FileNameb,namekey)
  SoSets *tN1;
  char *FileNameA,*FileNameb,*namekey;
{
  int i,j,nclusters,MinOpsLoc,nlb;
  int nOpsTotalGS,nOpsBlockGS,nOpsBorderGS;
  int nOpsTotal,MaxOps,MinOps;
  int totindep,minindep,maxindep,totblocks,totborder;
  int totindepGS,minindepGS,maxindepGS,totblockGS,totborderGS;
  char filenameA[40],filenameb[40];
  char filename[40];
  char *string = "SuperCluster";
  char *name;
  SoSElement *soselement;
  SetElement *element;
  Set *tN1i,*MaxOpsSet;
  SoSets *SC;

printf("PigeonHoleLoadBalance --- namekey=%s - nprocs=%d \n",namekey,nprocs);
fflush(stdout);

  if (loadbal != NULL) free(loadbal);
  loadbal = (LoadBalance *) calloc((nprocs+1),sizeof(LoadBalance));
  if (loadbal == NULL)
   {
    printf("can't allocate memory for loadbal[] \n");
    exit(1);
   } /* end if */

/* reset the SuperCluster values to NO */
/*
PrintSetOfSets(tN1);

printf("reset the SuperCluster values to NO \n");
fflush(stdout);
*/
  nOpsTotalGS = nOpsBlockGS = nOpsBorderGS = 0;
  soselement = tN1->First;
  while (soselement != NULL)
   {
    tN1i = soselement->set;
    tN1i->SuperCluster = NO;
    nOpsTotalGS  += tN1i->nOpsTotalGS;
    nOpsBlockGS  += tN1i->nOpsBlockGS;
    nOpsBorderGS += tN1i->nOpsBorderGS;
    nOpsTotal    += tN1i->nOpsTotal;
    soselement = soselement->Next;
   } /* end while soselement !=NULL */

printf("nOpsTotalGS=%d - nOpsBlockGS=%d - nOpsBorderGS=%d \n",
        nOpsTotalGS,nOpsBlockGS,nOpsBorderGS);
printf("  nOpsTotal=%d \n",nOpsTotal);
fflush(stdout);

/* load the first number of clusters into the pigeon holes */

printf("load the first number of clusters into the pigeon holes\n");
fflush(stdout);

  nclusters = tN1->n;
  if (nclusters < nprocs)
   {
    printf(" The number of clusters is less than the number of pigeon holes\n");
    printf("  number of clusters = %d  --  number of processors = %d \n",nclusters,nprocs);
    exit(1);
   } /* end if number of sets < nprocs */
  else
   {
    for (i=1; i<=nprocs; i++)
     {
      MaxOps = 0;
      soselement = tN1->First;
      while (soselement != NULL)
       {
        tN1i = soselement->set;
/*
printf("%s --- tN1i->nOpsTotal = %d --- ",tN1i->Name,tN1i->nOpsTotal);
fflush(stdout);
printf(" --- tN1i->SuperCluster = %d\n",tN1i->SuperCluster);
fflush(stdout);
*/
        if ( (tN1i->nOpsTotal > MaxOps) && (tN1i->SuperCluster == NO) )
         {
          MaxOps = tN1i->nOpsTotal;
          MaxOpsSet = tN1i;
/*
printf("MaxOps = %d -- ",MaxOps);
fflush(stdout);
printf(" -- MaxOpsSet = %s\n",tN1i->Name);
fflush(stdout);
*/
         } /* end if */
        soselement = soselement->Next;
       } /* end while soselement !=NULL */
      name = (char *) calloc(16,sizeof(char));
      if (name == NULL)
       {
        printf("can't allocate memory for name[] \n");
        exit(1);
       } /* end if */
      for (j=0; j<12; j++)
        name[j] = string[j];
      name[12] = '-';
      name[13] = (char) (((i%100)/10) + 48);
      name[14] = (char) ((i%10) + 48);
      name[15] = '\0';
      SC = NULL;
      SC = InitSetOfSets(SC,name);
      AddSetToSetOfSets(SC,MaxOpsSet);
      loadbal[i].SuperCluster = SC;
/*
PrintSetOfSets(loadbal[i].SuperCluster);
*/
      loadbal[i].nOpsBlockGS  = MaxOpsSet->nOpsBlockGS;
      loadbal[i].nOpsBorderGS = MaxOpsSet->nOpsBorderGS;
      loadbal[i].nOpsTotalGS  = MaxOpsSet->nOpsTotalGS;
      loadbal[i].nOpsTotal  = MaxOpsSet->nOpsTotal;
      MaxOpsSet->SuperCluster = i;
     } /* end for i */
   } /* end if number of sets >= nprocs */

/* place the remaining clusters into super clusters in a manner to
   balance the load - place the largest remaining cluster into the
   smallest super cluster - largest remaining cluster has the greatest 
   nOpsBlock value and smallest super block has the least total nOpsBlock */

printf("place the remaining clusters into super clusters\n");
fflush(stdout);

  for (i=1; i<=(nclusters-nprocs); i++)
   {
/*
printf("find the largest remaining cluster --- i=%d  \n",i);
fflush(stdout);
*/
/* find the largest remaining cluster */
    MaxOps = -1;
    MaxOpsSet = NULL;
    soselement = tN1->First;
    while (soselement != NULL)
     {
      tN1i = soselement->set;
      if ( (tN1i->nOpsTotal > MaxOps) && (tN1i->SuperCluster == NO) )
       {
        MaxOps = tN1i->nOpsTotal;
        MaxOpsSet = tN1i;
       } /* end if */
      soselement = soselement->Next;
     } /* end while soselement !=NULL */
/*
printf("\tfind the smallest super cluster \n");
fflush(stdout);
*/
/*
PrintSet(MaxOpsSet);
*/
/* find the smallest super cluster */
    MinOps = INFI;
    MinOpsLoc = 0;
    for (j=1; j<=nprocs; j++)
     {
      if (loadbal[j].nOpsTotal < MinOps)
       {
        MinOps = loadbal[j].nOpsTotal;
        MinOpsLoc = j;
       } /* end if */
     } /* end for j */
/*
printf("\tupdate the values \n");
fflush(stdout);
*/
/* update the values */
    AddSetToSetOfSets(loadbal[MinOpsLoc].SuperCluster,MaxOpsSet);
    loadbal[MinOpsLoc].nOpsBlockGS  += MaxOpsSet->nOpsBlockGS;
    loadbal[MinOpsLoc].nOpsBorderGS += MaxOpsSet->nOpsBorderGS;
    loadbal[MinOpsLoc].nOpsTotalGS  += MaxOpsSet->nOpsTotalGS;
    loadbal[MinOpsLoc].nOpsTotal      += MaxOpsSet->nOpsTotal;
    MaxOpsSet->SuperCluster = MinOpsLoc;
   } /* end for i */
/*
printf("\tprint out the super clusters \n");
fflush(stdout);
*/
/* print out the super clusters */
  totindepGS = totblockGS = totborderGS = 0;
  minindepGS = INFI;
  maxindepGS = 0;
  totindep = 0;
  minindep = INFI;
  maxindep = 0;
  for (i=1; i<=nprocs; i++)
   {

    printf("\nSuper Cluster Number %d --- for nprocs=%d \n",i,nprocs);
    printf("\tNumber of operations (GS = %d \n",loadbal[i].nOpsTotalGS);
    printf("\tTotal number of operations = %d \n",loadbal[i].nOpsTotal);
    fflush(stdout);

/*
    PrintSetOfSets(loadbal[i].SuperCluster);
*/
    totindepGS  += loadbal[i].nOpsTotalGS;
    totblockGS  += loadbal[i].nOpsBlockGS;
    totborderGS += loadbal[i].nOpsBorderGS;
    if (loadbal[i].nOpsTotalGS < minindepGS) minindepGS = loadbal[i].nOpsTotalGS;
    if (loadbal[i].nOpsTotalGS > maxindepGS) maxindepGS = loadbal[i].nOpsTotalGS;
    totindep  += loadbal[i].nOpsTotal;
    if (loadbal[i].nOpsTotal < minindep) minindep = loadbal[i].nOpsTotal;
    if (loadbal[i].nOpsTotal > maxindep) maxindep = loadbal[i].nOpsTotal;
   } /* end for i */

  nlb = N2->n;
  printf("\nSummary Statistics for nprocs=%d \n",nprocs);
  printf("\tTotal number of operations in the independent blocks = %lf \n",totindep);
  printf("\t\tTotal number of operations in the independent blocks (GS) = %d \n",totindepGS);
  printf("\t\tTotal number of operations in the blocks             (GS) = %d \n",totblockGS);
  printf("\t\tTotal number of operations in the borders            (GS) = %d \n\n",totborderGS);
  printf("\tTotal number of operations in the lastblock (sparse GS)     = %d \n",nOpsLastBlockGS);
  printf("\tMinimum number of operations in a cluster            = %lf \n",minindep);
  printf("\tMaximum number of operations in a cluster            = %lf \n",maxindep);
  printf("\t(Maximum-Minimum) operations in a cluster            = %lf \n\n",(maxindep - minindep));
  printf("\tPercent of operations in the clusters (sparse)       = %f \n",
                     (double) ((double) (totindep * 100) / (double) (totindep + nOpsLastBlockGS)));


/* place the superclusters in contigious locations */
/* initialize the set of sets sN1 */
  sN1 = NULL;
  sN1 = InitSetOfSets(sN1,"SuperClusters");
  if (proc != NULL) free(proc);
  proc = (Proc *) calloc((nprocs+1),sizeof(Proc));
  if (proc == NULL)
   {
    printf("can't allocate memory for proc[] \n");
    exit(1);
   } /* end if */

/* copy the sets to the set of set sN1 so the superclusters
   are in adjacent order */
   proc[0].BlockPtr          = 0;
   proc[0].DiagPtr           = 0;
   proc[0].BlockRowPtr       = 0;
   proc[0].LowerBorderPtr    = 0;
   proc[0].LowerBorderRowPtr = 0;
   proc[i].LBSet             = NULL;
   proc[i].LBSet             = InitSet(proc[i].LBSet,"LastBlockSet");
  for (i=nprocs; i>=1; i--)
    {
     proc[i].BlockPtr          = 0;
     proc[i].DiagPtr           = 0;
     proc[i].BlockRowPtr       = 0;
     proc[i].LowerBorderPtr    = 0;
     proc[i].LowerBorderRowPtr = 0;
     proc[i].LBSet             = NULL;
     proc[i].LBSet             = InitSet(proc[i].LBSet,"LastBlockSet");

     SC = loadbal[i].SuperCluster;

printf("(loadbal[%d].SuperCluster)->Name=%s - (loadbal[%d].SuperCluster)->n=%d \n",
         i,(loadbal[i].SuperCluster)->Name,i,(loadbal[i].SuperCluster)->n);
fflush(stdout);

     soselement = SC->First;
     while (soselement != NULL)
      {
       tN1i = soselement->set;
       AddSetToSetOfSets(sN1,tN1i);
       proc[i].BlockPtr       += 1;
       proc[i].DiagPtr        += tN1i->n;

printf("\ttN1i->Name=%s - n=%d - nOTGS=%d  \n",
               tN1i->Name,tN1i->n,tN1i->nOpsTotalGS);
fflush(stdout);

       soselement = soselement->Next;
      } /* soselement != NULL */
    } /* end for i */
/* sum values to make proc[].BlockPtr contigious */

printf("\tproc[0].BlockPtr=%d \n",proc[0].BlockPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].BlockPtr += proc[i-1].BlockPtr;

printf("\tproc[%d].BlockPtr=%d \n",i,proc[i].BlockPtr);
fflush(stdout);

    } /* end for i */
/* sum values to make proc[].DiagPtr contigious */

printf("\tproc[0].DiagPtr=%d \n",proc[0].DiagPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].DiagPtr += proc[i-1].DiagPtr;

printf("\tproc[%d].DiagPtr=%d \n",i,proc[i].DiagPtr);
fflush(stdout);

    } /* end for i */
/* reset the adjacency list */
  ResetAdjList();
/*
PrintSetOfSets(sN1);
*/
/* call ReorderNetworkClusters() to generate a new adjlist */

printf("ReorderNetworkClusters() \n");
fflush(stdout);

  ReorderNetworkClusters(sN1,namekey);

printf("GenerateBitmap \n");
fflush(stdout);
  GenerateBitmap(MakeOutputFilename(namekey,filename,(100+nprocs))); /* produce a bitmap of the graph */

printf("PrintL \n");
fflush(stdout);
  PrintL("NiMo Data"); /* produce an X Window display of the graph */

/* generate data for bdb-form GS solver */

printf("BDBFormData() \n");
fflush(stdout);

  BDBFormData();
/* write data for bdb-for GS solver to a file */

printf("OutputBDBFormData() \n");
fflush(stdout);

  OutputBDBFormData(MakeOutputFilename(FileNameA,filenameA,nprocs),
                    MakeOutputFilename(FileNameb,filenameb,nprocs));

} /* end PigeonHoleLoadBalance() */

/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void BDBFormData()
{
  int i,j,k,l,m,done,iproc,jproc,icolor,here;
  int nBD,nLB,nBor;
  int nlastblockstart,nlastblockend;
  Set *N1i;
  SoSElement *soselement;
  SetElement *element;

if (nprocs == 1)
 {
  nblocks = 1;

printf("nblocks=%d  \n",nblocks);
fflush(stdout);

  if (blocks != NULL) free(blocks);
  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    exit(1);
   } /* end if */
/*---------------------------------------------------*/
printf("determine blocks --- \n");
fflush(stdout);

  blocks[0].DiagPtr = 0;
  blocks[1].DiagPtr = n;
  proc[1].BlockPtr  = 1;
  proc[1].DiagPtr   = n;

printf("BDBFormData() --- i=%d - nblocks=%d \n",i,nblocks);
fflush(stdout);

/*---------------------------------------------------*/
printf("determine diag --- \n");
fflush(stdout);

  ndiag = n;

printf("\t\tndiag=%d  \n",ndiag);
fflush(stdout);

  if (diag != NULL) free(diag);
  diag = (Diag *) calloc((ndiag+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    exit(1);
   } /* end if */
  for (i=1;i<=ndiag;i++)
    {
     diag[i].Value = ndiag + drand48();
     diag[i].BlockRowSet  = NULL;
     diag[i].BlockRowSet  = InitSet(diag[i].BlockRowSet,"BlockRowSet");
    } /* end for i */
/*---------------------------------------------------*/

printf("determine elements in BlockRowSet  \n");
fflush(stdout);

 for (k=1; k<=nadjlist; k++) /* for all non-zero elements */
   {
     i = adjlist[k].r.NodeNum;
     j = adjlist[k].l.NodeNum;
     AddElementToSet(diag[i].BlockRowSet,j);
     AddElementToSet(diag[j].BlockRowSet,i);
   } /* end for k */

  nBD = nadjlist;
  nLB = nBor = 0;

printf("\tnBD=%d - nLB=%d - nBor=%d - \n",nBD,nLB,nBor);
fflush(stdout);

/*---------------------------------------------------*/
  proc[1].LowerBorderPtr = 0;
/*---------------------------------------------------*/

printf("determine diag[].PtrAix --- \n");
fflush(stdout);

  diag[0].PtrAix = 0;
  for (i=1;i<=n;i++)
    {
     diag[i].PtrAix = diag[i-1].PtrAix + (diag[i].BlockRowSet)->n;
    } /* end for i */
  proc[1].BlockRowPtr = 2 * nadjlist;

printf("\tproc[0].BlockRowPtr=%d \n",proc[0].BlockRowPtr);
printf("\tproc[1].BlockRowPtr=%d \n",proc[1].BlockRowPtr);
fflush(stdout);
/*---------------------------------------------------*/

printf("determine blockrow --- \n");
fflush(stdout);

  nblockrow = 0;
  for (i=1;i<=n;i++)
    nblockrow += (diag[i].BlockRowSet)->n;

printf("\t\t --- nblockrow=%d \n",nblockrow);
fflush(stdout);

  if (blockrow != NULL) free(blockrow);
  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    exit(1);
   } /* end if */
  k = 1;

printf("\t\t --- nblockrow=%d - k=%d \n",nblockrow,k);
fflush(stdout);

  for (i=1;i<=n;i++)
    {

printf("\t\t --- k=%d - i=%d - (diag[%d].BlockRowSet)->n=%d - nblockrow=%d \n",
          k,i,i,(diag[i].BlockRowSet)->n,nblockrow);
fflush(stdout);
     while ((diag[i].BlockRowSet)->n > 0)
       {
        blockrow[k].Value = -1.0 * drand48();
        blockrow[k].i = i;
        blockrow[k].j = MinValueInSet(diag[i].BlockRowSet);
/*
printf("\t\t --- blockrow[%d].Value=%lf - blockrow[%d].i=%d - blockrow[%d].j=%d \n",
              k,blockrow[k].Value,k,blockrow[k].i,k,blockrow[k].j);
fflush(stdout);
*/
        k++;
       } /* end while */
    } /* end for i */

printf("\t\t --- nblockrow=%d - k=%d \n",nblockrow,k);
fflush(stdout);
/*---------------------------------------------------*/

printf("determine lborder --- \n");
fflush(stdout);

  nlborder = 0;

printf("\t\t --- nlborder=%d  \n",nlborder);
fflush(stdout);

  if (lborder != NULL) free(lborder);
  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    exit(1);
   } /* end if */

  nlborderrow = 0;

printf("\t\t --- nlborderrow=%d  \n",nlborderrow);
fflush(stdout);

  if (lborderrow != NULL) free(lborderrow);
  lborderrow = (LBorderRow *) calloc((nlborderrow+1),sizeof(LBorderRow));
  if (lborderrow == NULL)
   {
    printf("can't allocate memory for lborderrow[] \n");
    exit(1);
   } /* end if */

  proc[1].LowerBorderRowPtr = 0;
printf("\tproc[0].LowerBorderRowPtr=%d \n",proc[0].LowerBorderRowPtr);
printf("\tproc[1].LowerBorderRowPtr=%d \n",proc[1].LowerBorderRowPtr);
fflush(stdout);

/*---------------------------------------------------*/
printf("determine values in lbcolor[]. --- \n");
fflush(stdout);
  ncolors = 0;
  if (lbcolortmp != NULL) free(lbcolortmp);
  lbcolortmp = (LastBlockColor *) calloc(((nprocs+1)*(ncolors+1)),sizeof(LastBlockColor));
  if (lbcolortmp == NULL)
   {
    printf("can't allocate memory for lbcolortmp[][] \n");
    exit(1);
   } /* end if */
  if (lbcolor != NULL) free(lbcolor);
  lbcolor = (LastBlockColor **) calloc((ncolors+1),sizeof(LastBlockColor *));
  if (lbcolor == NULL)
   {
    printf("can't allocate memory for lbcolor[][] \n");
    exit(1);
   } /* end if */
  for (icolor=0;icolor<=ncolors;icolor++) lbcolor[icolor] = lbcolortmp + (icolor*(nprocs+1));
/*---------------------------------------------------*/
printf("determine ldiag --- \n");
fflush(stdout);

  nldiag = 0;

printf("\t\tnldiag=%d  \n",nldiag);
fflush(stdout);

  if (ldiag != NULL) free(ldiag);
  ldiag = (LastDiag *) calloc((nldiag+1),sizeof(LastDiag));
  if (ldiag == NULL)
   {
    printf("can't allocate memory for ldiag[] \n");
    exit(1);
   } /* end if */
/*---------------------------------------------------*/
printf("determine lblockrow --- \n");
fflush(stdout);

  nlblockrow = 0;

printf("\t\t --- nlblockrow=%d \n",nlblockrow);
fflush(stdout);

  if (lblockrow != NULL) free(lblockrow);
  lblockrow = (LastBlockRow *) calloc((nlblockrow+1),sizeof(LastBlockRow));
  if (lblockrow == NULL)
   {
    printf("can't allocate memory for lblockrow[] \n");
    exit(1);
   } /* end if */

/*---------------------------------------------------*/
/* do a check on sets to see that all values are accounted for*/
  k = 0;
  for (i=1;i<=ndiag;i++)
    {
     k += (diag[i].BlockRowSet)->n;
    } /* end for i */
  printf("Check on set diag[].BlockRowSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

 } /* ----------end if nprocs == 1 -------------------------------*/
else
 {
  nlastblockstart = n - (N2->n) + 1;
  nlastblockend   = n;

printf("nlastblockstart=%d - nlastblockend=%d \n",nlastblockstart,nlastblockend);
fflush(stdout);

/* determine the size of the data-structure elements */
/* allocate memory for the bdb-form sparse matrix data structures */
  nblocks = sN1->n;

printf("nblocks=%d  \n",nblocks);
fflush(stdout);

  if (blocks != NULL) free(blocks);
  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    exit(1);
   } /* end if */
/*---------------------------------------------------*/
printf("determine blocks --- \n");
fflush(stdout);

  blocks[0].DiagPtr = 0;
  i = 1;
  soselement = sN1->First;
  while (soselement != NULL)
    {
     blocks[i].BlockID = i;
     N1i = soselement->set;
     blocks[i].BlockSz = N1i->n;
     blocks[i].DiagPtr = blocks[i-1].DiagPtr + N1i->n;
     for (j=1;j<=nprocs;j++)
       {
        if ( (blocks[i-1].DiagPtr >= proc[j-1].DiagPtr) && 
             (blocks[i].DiagPtr   <= proc[j].DiagPtr) )
	  {
           blocks[i].ProcNum = (j-1);
           break;
	  } /* end if */
       } /* end for l*/
     soselement = soselement->Next;
     i++;
    } /* soselement != NULL */

printf("BDBFormData() --- i=%d - nblocks=%d \n",i,nblocks);
fflush(stdout);

/*---------------------------------------------------*/
printf("determine diag --- \n");
fflush(stdout);

  ndiag = n;

printf("\t\tndiag=%d  \n",ndiag);
fflush(stdout);

  if (diag != NULL) free(diag);
  diag = (Diag *) calloc((ndiag+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    exit(1);
   } /* end if */
  for (i=1;i<=ndiag;i++)
    {
     diag[i].Value = ndiag + drand48();
     diag[i].BlockRowSet  = NULL;
     diag[i].BlockRowSet  = InitSet(diag[i].BlockRowSet,"BlockRowSet");
    } /* end for i */
/*---------------------------------------------------*/

printf("determine elements in BlockRowSet  \n");
fflush(stdout);

 for (k=1; k<=nadjlist; k++) /* for all non-zero elements */
   {
    if ( ( (adjlist[k].l.NodeNum <  nlastblockstart) && (adjlist[k].r.NodeNum <  nlastblockstart) ) ||
         ( (adjlist[k].l.NodeNum >= nlastblockstart) && (adjlist[k].r.NodeNum >= nlastblockstart) ) )
      {
	i = adjlist[k].r.NodeNum;
        j = adjlist[k].l.NodeNum;
	AddElementToSet(diag[i].BlockRowSet,j);
        AddElementToSet(diag[j].BlockRowSet,i);
      } /* end if in diagonal-bordered blocks or last block */
    else
      {
       if (adjlist[k].l.NodeNum < adjlist[k].r.NodeNum)
         {
	  i = adjlist[k].r.NodeNum;
          j = adjlist[k].l.NodeNum;
	 } /* end if */
       else
	 {
	  i = adjlist[k].l.NodeNum;
          j = adjlist[k].r.NodeNum;
	 } /* end else */
       AddElementToSet(diag[j].BlockRowSet,i);
       for (l=1;l<=nprocs;l++)
	{
         if ( (j > proc[l-1].DiagPtr) && (j <= proc[l].DiagPtr) )
	   {
            AddElementToSet(proc[l].LBSet,i);
/*
printf("j=%d --- AddElementToSet(procs[%d].LBSet,%d) \n",j,l,i);
fflush(stdout);
*/
            break;
	   } /* end if j is in a lower block */
	} /* end for l */
      } /* end if in lower border */
   } /* end for k */
/*---------------------------------------------------*/
 nBD = nLB = nBor = 0;
 for (k=1; k<=nadjlist; k++) /* for all non-zero elements */
   {
    if ( (adjlist[k].l.NodeNum <  nlastblockstart) && (adjlist[k].r.NodeNum <  nlastblockstart) )
      {
        nBD++;
      } /* end if in diagonal-bordered blocks */
    else if ( (adjlist[k].l.NodeNum >= nlastblockstart) && (adjlist[k].r.NodeNum >= nlastblockstart) ) 
      {
        nLB++;
      } /* end if in last block */
    else
      {
        nBor++;
      } /* end if in lower border */
   } /* end for k */

printf("\tnBD=%d - nLB=%d - nBor=%d - \n",nBD,nLB,nBor);
fflush(stdout);

/*---------------------------------------------------*/

printf("determine proc[].LowerBorderPtr --- \n");
fflush(stdout);

  proc[0].LowerBorderPtr = 0;

printf("\tproc[0].LowerBorderPtr=%d \n",proc[0].LowerBorderPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].LowerBorderPtr = proc[i-1].LowerBorderPtr + (proc[i].LBSet)->n;

printf("\tproc[%d].LowerBorderPtr=%d \n",i,proc[i].LowerBorderPtr);
fflush(stdout);

    } /* end for i */
/*---------------------------------------------------*/

printf("determine diag[].PtrAix --- \n");
fflush(stdout);

  diag[0].PtrAix = 0;
  for (i=1;i<nlastblockstart;i++)
    {
     diag[i].PtrAix = diag[i-1].PtrAix + (diag[i].BlockRowSet)->n;
     for (iproc=1;iproc<=nprocs;iproc++)
       {
        if ( (i > proc[iproc-1].DiagPtr) && (i <= proc[iproc].DiagPtr) )
	  {
           proc[iproc].BlockRowPtr += (diag[i].BlockRowSet)->n;
           break;
	  } /* end if */
       } /* end for iproc */
    } /* end for i */
/* sum values to make proc[].BlockRowPtr contigious */

printf("\tproc[0].BlockRowPtr=%d \n",proc[0].BlockRowPtr);
fflush(stdout);

   for (iproc=1;iproc<=nprocs;iproc++)
    {
     proc[iproc].BlockRowPtr += proc[iproc-1].BlockRowPtr;

printf("\tproc[%d].BlockRowPtr=%d \n",iproc,proc[iproc].BlockRowPtr);
fflush(stdout);

    } /* end for iproc */
/*---------------------------------------------------*/

printf("determine blockrow --- \n");
fflush(stdout);

  nblockrow = 0;
  for (i=1;i<nlastblockstart;i++)
    nblockrow += (diag[i].BlockRowSet)->n;

printf("\t\t --- nblockrow=%d \n",nblockrow);
fflush(stdout);

  if (blockrow != NULL) free(blockrow);
  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    exit(1);
   } /* end if */
  k = 1;
/*
printf("\t\t --- nblockrow=%d - k=%d \n",nblockrow,k);
fflush(stdout);
*/
  for (i=1;i<nlastblockstart;i++)
    {
/*
printf("\t\t --- k=%d - i=%d - (diag[%d].BlockRowSet)->n=%d - nblockrow=%d \n",
          k,i,i,(diag[i].BlockRowSet)->n,nblockrow);
fflush(stdout);
*/
     while ((diag[i].BlockRowSet)->n > 0)
       {
        blockrow[k].Value = -1.0 * drand48();
        blockrow[k].i = i;
        blockrow[k].j = MinValueInSet(diag[i].BlockRowSet);
/*
printf("\t\t --- blockrow[%d].Value=%lf - blockrow[%d].i=%d - blockrow[%d].j=%d \n",
              k,blockrow[k].Value,k,blockrow[k].i,k,blockrow[k].j);
fflush(stdout);
*/
        k++;
       } /* end while */
    } /* end for i */

printf("\t\t --- nblockrow=%d - k=%d \n",nblockrow,k);
fflush(stdout);
/*---------------------------------------------------*/

printf("determine lborder --- \n");
fflush(stdout);

  nlborder = 0;
  for (i=1;i<=nprocs;i++)
    nlborder += (proc[i].LBSet)->n;

printf("\t\t --- nlborder=%d  \n",nlborder);
fflush(stdout);

  if (lborder != NULL) free(lborder);
  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    exit(1);
   } /* end if */

printf("initialize LBorderRowSet --- \n");
fflush(stdout);

  k = 1;
  for (i=1;i<=nprocs;i++)
    {
     while ((proc[i].LBSet)->n > 0)
       {
        lborder[k].i = MinValueInSet(proc[i].LBSet);
        lborder[k].LBorderRowSet = NULL;
        lborder[k].LBorderRowSet = InitSet(lborder[k].LBorderRowSet,"LBorderRowSet");
        k++;
       } /* end while */
    } /* end for i */
/*---------------------------------------------------*/

printf("determine elements of LBorderRowSet --- \n");
fflush(stdout);
 
 for (k=1; k<=nadjlist; k++) /* for all non-zero elements */
   {
    done = NO;
    if ( ( (adjlist[k].l.NodeNum < nlastblockstart) && (adjlist[k].r.NodeNum >= nlastblockstart) ) ||
         ( (adjlist[k].l.NodeNum >= nlastblockstart) && (adjlist[k].r.NodeNum < nlastblockstart) ) )
      {
      if ( (adjlist[k].l.NodeNum < nlastblockstart) && (adjlist[k].r.NodeNum >= nlastblockstart) )
        {
         i = adjlist[k].r.NodeNum;
         j = adjlist[k].l.NodeNum;
        } /* end if in border */
      else if ( (adjlist[k].l.NodeNum >= nlastblockstart) && (adjlist[k].r.NodeNum < nlastblockstart) )
        {
         i = adjlist[k].l.NodeNum;
         j = adjlist[k].r.NodeNum;
        } /* end if in border */
       for(l=1;( (l<=nprocs) && (done == NO) );l++) /* find the processor */
         {
          if ( (j > proc[l-1].DiagPtr) && (j <= proc[l].DiagPtr))
	    {
/* find the location in lborder[] */
             for (m=(proc[l-1].LowerBorderPtr+1);( (m<=proc[l].LowerBorderPtr) && (done == NO) );m++)
	      {
               if (lborder[m].i == i)
	         {
                  AddElementToSet(lborder[m].LBorderRowSet,j);
                  done = YES;
	         } /* end if in lborder[m] */
	      }   /* end for m */
	    } /* end if in proc l */
         } /* end for l */
      } /* end if */
   } /* end for k */

  nlborderrow = 0;
  for (i=1;i<=nlborder;i++)
   nlborderrow += (lborder[i].LBorderRowSet)->n;

printf("\t\t --- nlborderrow=%d  \n",nlborderrow);
fflush(stdout);

  if (lborderrow != NULL) free(lborderrow);
  lborderrow = (LBorderRow *) calloc((nlborderrow+1),sizeof(LBorderRow));
  if (lborderrow == NULL)
   {
    printf("can't allocate memory for lborderrow[] \n");
    exit(1);
   } /* end if */
/*---------------------------------------------------*/

printf("determine values of lborderrow --- \n");
fflush(stdout);

  for (i=1;i<=nlborder;i++)
   {
     for (j=1;j<=nprocs;j++)
       {
        if (( i > proc[j-1].LowerBorderPtr) && (i <= proc[j].LowerBorderPtr) )
	  {
           proc[j].LowerBorderRowPtr += (lborder[i].LBorderRowSet)->n;
           break;
	  } /* end if */
       } /* end for j */
   } /* end for i */
 /* sum values to make proc[].LowerBorderRowPtr contigious */

printf("\tproc[0].LowerBorderRowPtr=%d \n",proc[0].LowerBorderRowPtr);
fflush(stdout);

  for (i=1;i<=nprocs;i++)
    {
     proc[i].LowerBorderRowPtr += proc[i-1].LowerBorderRowPtr;

printf("\tproc[%d].LowerBorderRowPtr=%d \n",i,proc[i].LowerBorderRowPtr);
fflush(stdout);

    } /* end for i */
 
  lborder[0].PtrLBix = 0;
  k = 1;
  for (i=1;i<=nlborder;i++)
   {
    lborder[i].PtrLBix = lborder[i-1].PtrLBix + (lborder[i].LBorderRowSet)->n;
/*
printf("\t\tk=%d - i=%d - lborder[%d].i=%d - lborder[%d].PtrLBix=%d - (lborder[%d].LBorderRowSet)->n=%d \n",
             k,i,i,lborder[i].i,i,lborder[i].PtrLBix,i,(lborder[i].LBorderRowSet)->n);
fflush(stdout);
*/
    while ((lborder[i].LBorderRowSet)->n > 0)
      {
        lborderrow[k].Value = -1.0 * drand48();
        lborderrow[k].i = lborder[i].i;
        lborderrow[k].j = MinValueInSet(lborder[i].LBorderRowSet);
/*
printf("\t\ti=%d - k=%d - lborder[%d].i=%d - lborderrow[%d].Value=%lf - lborderrow[%d].i=%d - lborderrow[%d].j=%d \n",
             i,k,i,lborder[i].i,k,lborderrow[k].Value,k,lborderrow[k].i,k,lborderrow[k].j);
fflush(stdout);
*/
        k++;
       } /* end while */
   } /* end for i */
/*---------------------------------------------------*/

printf("determine values in lbcolor[]. --- \n");
fflush(stdout);

  if (lbcolortmp != NULL) free(lbcolortmp);
  lbcolortmp = (LastBlockColor *) calloc(((nprocs+1)*(ncolors+1)),sizeof(LastBlockColor));
  if (lbcolortmp == NULL)
   {
    printf("can't allocate memory for lbcolortmp[][] \n");
    exit(1);
   } /* end if */
  if (lbcolor != NULL) free(lbcolor);
  lbcolor = (LastBlockColor **) calloc((ncolors+1),sizeof(LastBlockColor *));
  if (lbcolor == NULL)
   {
    printf("can't allocate memory for lbcolor[][] \n");
    exit(1);
   } /* end if */
  for (icolor=0;icolor<=ncolors;icolor++) lbcolor[icolor] = lbcolortmp + (icolor*(nprocs+1));
/*---------------------------------------------------*/
  for (iproc=1;iproc<=nprocs;iproc++)
    {
     lbcolor[0][iproc].Color   = 0;
     lbcolor[0][iproc].ProcNum = 0;
     lbcolor[0][iproc].Starti  = 0;
     lbcolor[0][iproc].DiagPtr = 0;
    } /* end for iproc */ 
  for (icolor=1;icolor<=ncolors;icolor++)
    {
     lbcolor[icolor][0].Color   = 0;
     lbcolor[icolor][0].ProcNum = 0;
     lbcolor[icolor][0].Starti  = 0;
     lbcolor[icolor][0].DiagPtr = lbcolor[icolor-1][nprocs].DiagPtr;
     for (iproc=1;iproc<nprocs;iproc++)
       {
        lbcolor[icolor][iproc].Color   = icolor;
        lbcolor[icolor][iproc].ProcNum = iproc;
        lbcolor[icolor][iproc].Starti  = commlb[icolor][iproc].starti;
        lbcolor[icolor][iproc].DiagPtr = (commlb[icolor][iproc+1].starti - 1);
       } /* end for iproc */ 
      lbcolor[icolor][nprocs].Color   = icolor;
      lbcolor[icolor][nprocs].ProcNum = iproc;
      lbcolor[icolor][nprocs].Starti  = commlb[icolor][iproc].starti;
      if (icolor < ncolors)
        lbcolor[icolor][nprocs].DiagPtr = (commlb[icolor+1][1].starti - 1);
      else
        lbcolor[icolor][nprocs].DiagPtr = (N2->n);
    } /* end for icolor */
/*---------------------------------------------------*/
printf("determine ldiag --- \n");
fflush(stdout);

  nldiag = (N2->n);

printf("\t\tnldiag=%d  \n",nldiag);
fflush(stdout);

  if (ldiag != NULL) free(ldiag);
  ldiag = (LastDiag *) calloc((nldiag+1),sizeof(LastDiag));
  if (ldiag == NULL)
   {
    printf("can't allocate memory for ldiag[] \n");
    exit(1);
   } /* end if */
  ldiag[0].PtrAix = 0;
  for (i=1;i<=nldiag;i++)
    {
     j = (i + nlastblockstart - 1);
     ldiag[i].Value = ndiag + drand48();
     ldiag[i].PtrAix = ldiag[i-1].PtrAix + (diag[j].BlockRowSet)->n;
    } /* end for i */
/*---------------------------------------------------*/
printf("determine lblockrow --- \n");
fflush(stdout);

  nlblockrow = (nadjlistLB * 2);

printf("\t\t --- nlblockrow=%d \n",nlblockrow);
fflush(stdout);

  if (lblockrow != NULL) free(lblockrow);
  lblockrow = (LastBlockRow *) calloc((nlblockrow+1),sizeof(LastBlockRow));
  if (lblockrow == NULL)
   {
    printf("can't allocate memory for lblockrow[] \n");
    exit(1);
   } /* end if */
  k = 1;
/*
printf("\t\t --- nlblockrow=%d - k=%d \n",nlblockrow,k);
fflush(stdout);
*/
  for (i=nlastblockstart;i<=nlastblockend;i++)
    {
/*
printf("\t\t --- k=%d - i=%d - (diag[%d].BlockRowSet)->n=%d - nlblockrow=%d \n",
          k,i,i,(diag[i].BlockRowSet)->n,nblockrow);
fflush(stdout);
*/
     while ((diag[i].BlockRowSet)->n > 0)
       {
        lblockrow[k].Value = -1.0 * drand48();
        lblockrow[k].i = i;
        lblockrow[k].j = MinValueInSet(diag[i].BlockRowSet);
/*
printf("\t\t --- lblockrow[%d].Value=%lf - lblockrow[%d].i=%d - lblockrow[%d].j=%d \n",
              k,lblockrow[k].Value,k,lblockrow[k].i,k,lblockrow[k].j);
fflush(stdout);
*/
        k++;
       } /* end while */
    } /* end for i */

printf("\t\t --- nlblockrow=%d - k=%d \n",nlblockrow,k);
fflush(stdout);
/*---------------------------------------------------*/
/* calculate lborder[].ulbproc and lborder[].ulbloc */
  lborder[0].ulbproc = 0;
  lborder[0].ulbloc  = 0;
  for (iproc=1;iproc<=nprocs;iproc++) /* source processor */
   {
    for (jproc=1;jproc<=nprocs;jproc++) /* destination processor */
      {
       k = 0;
       for (i=(proc[iproc-1].LowerBorderPtr + 1);i<=proc[iproc].LowerBorderPtr;i++)
	 {
           for (icolor=1;icolor<=ncolors;icolor++)
             {
              if ( ((lborder[i].i - nlastblockstart + 1) >  lbcolor[icolor][jproc-1].DiagPtr) && 
                   ((lborder[i].i - nlastblockstart + 1) <= lbcolor[icolor][jproc].DiagPtr) )
	        {
                 lborder[i].ulbproc = jproc;
                 here = NO;
                 for (j=(proc[iproc-1].LowerBorderPtr + 1);j<i;j++)
	           {
                    if ( (lborder[i].i == lborder[j].i) && (lborder[i].ulbproc == lborder[j].ulbproc) )
	             {
                      lborder[i].ulbloc = lborder[j].ulbloc;
                      here = YES;
                      break;
	             } /* end if */
	           } /* end for j */
                 if (here == NO)
	           {
/*
printf("iproc=%d - jproc=%d - i=%d - icolor=%d - k=%d  \n",iproc,jproc,i,icolor,k);
fflush(stdout);
*/
                    lborder[i].ulbloc = ++k;
	           } /* end if */
	        } /* end if */
             } /* end for icolor */
	 } /* end for i */
        ulbbufsize[iproc][jproc] = k;
/*
printf("\tiproc=%d - jproc=%d - sizeof(UpdateLastBlock)=%d - k=%d - ulbbufsize[%d][%d]=%d  \n",
         iproc,jproc,sizeof(UpdateLastBlock),k,iproc,jproc,ulbbufsize[iproc][jproc]);
fflush(stdout);
*/
      } /* end for jproc */   
   } /* end for iproc */

/*---------------------------------------------------*/
/* do a check on sets to see that all values are accounted for*/
  k = 0;
  for (i=1;i<=nprocs;i++)
    {
     k += (proc[i].LBSet)->n;
    } /* end for i */
  printf("Check on set proc[].LBSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

  k = 0;
  for (i=1;i<=ndiag;i++)
    {
     k += (diag[i].BlockRowSet)->n;
    } /* end for i */
  printf("Check on set diag[].BlockRowSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

  k = 0;
  for (i=1;i<=nlborder;i++)
    {
     k += (lborder[i].LBorderRowSet)->n;
    } /* end for i */
  printf("Check on set lborder[].LBorderRowSet --- remaining non-zero elements = %d \n",k);
  fflush(stdout);

 }/* ----------end if nprocs != 1 -------------------------------*/
} /* end BDBFormData() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void OutputBDBFormData(FileNameA,FileNameb)
  char *FileNameA,*FileNameb;
{
 int i,j,ndb,nldiag;
 int nlastblockstart,nlastblockend;
 FILE *fd;
printf("FileNameA =|%s| \n",FileNameA);
printf("FileNameb =|%s| \n",FileNameb);
fflush(stdout);
 fd = fopen(FileNameA,"w");
 
printf("write out the assignment of contigious diagonal elements to a processor nprocs=%d \n",nprocs);
fflush(stdout);
 
/* (0) write out the assignment of contigious diagonal elements to a processor */
  fprintf(fd," \t%d\n",nprocs);
  for (i=1; i<=nprocs; i++)
   {
    fprintf(fd,"%d \t%d %d %d %d %d %d \n",i,(i-1),
                proc[i].BlockPtr,
                proc[i].DiagPtr,
                proc[i].BlockRowPtr,
                proc[i].LowerBorderPtr,
                proc[i].LowerBorderRowPtr);
   } /* end for i */ 
 
printf("write out the block identifer array of structures \n");
fflush(stdout);
 
/* (1) write out the block identifer array of structures */
  fprintf(fd," \t%d\n",nblocks);
  for (i=1; i<=nblocks; i++)
   {
    fprintf(fd,"%d \t%d %d %d %d \n",i,
                blocks[i].BlockID,
                blocks[i].BlockSz,
                blocks[i].ProcNum,
                blocks[i].DiagPtr);
   } /* end for i */
 
printf("write out the matrix diagonal array of structures \n");
fflush(stdout);
 
/* (2) write out the matrix diagonal array of structures */
  if (nprocs == 1)
    ndb = n;
  else
    ndb = ndiag - (N2->n);
  fprintf(fd," \t%d\n",ndb);
  for (i=1; i<=ndb; i++)
   {
    fprintf(fd,"%d \t%lf %d \n",i,
                diag[i].Value,
                diag[i].PtrAix);
   } /* end for i */
 
printf("write out the block row storage array of structures \n");
fflush(stdout);
 
/* (3) write out the block row storage array of structures */
  fprintf(fd," \t%d\n",nblockrow);
  for (i=1; i<=nblockrow; i++)
   {
    fprintf(fd,"%d \t%lf %d %d \n",i,
                blockrow[i].Value,
                blockrow[i].i,
                blockrow[i].j );
   } /* end for i */
 
printf("write out the lower border row storage array of structures \n");
fflush(stdout);
 
/* (4) write out the lower border row storage array of structures */
  fprintf(fd," \t%d\n",nlborder);
  for (i=1; i<=nlborder; i++)
   {
    fprintf(fd,"%d \t%d %d %d %d \n",i,
                lborder[i].i,
                lborder[i].ulbproc,
                lborder[i].ulbloc,
                lborder[i].PtrLBix );
   } /* end for i */
 
printf("write out the lower border row storage array of structures \n");
fflush(stdout);
 
/* (5) write out the lower border row storage array of structures */
  fprintf(fd," \t%d\n",nlborderrow);
  for (i=1; i<=nlborderrow; i++)
   {
    fprintf(fd,"%d \t%lf %d %d \n",i,
                lborderrow[i].Value,
                lborderrow[i].i,
                lborderrow[i].j );
   } /* end for i */

printf("write-out the last diagonal block colors\n");
fflush(stdout);
 
/* (7) write out the last diagonal block colors identifer array of structures */
  fprintf(fd,"\t%d \n",ncolors);
  for (j=1; j<=ncolors; j++)
    {
     for (i=1; i<=nprocs; i++)
      {
       fprintf(fd,"%d \t%d %d %d %d \n",i,
                   lbcolor[j][i].Color,
                   lbcolor[j][i].ProcNum,
                   lbcolor[j][i].Starti,
                   lbcolor[j][i].DiagPtr );
      } /* end for i */
    } /* end for j */
 
printf("write out the last diagonal array of structures \n");
fflush(stdout);
 
/* (8) write out the last diagonal array of structures */
  if (nprocs == 1)
    nldiag = 0;
  else
    nldiag = (N2->n);
  fprintf(fd," \t%d\n",nldiag);
  for (i=1; i<=nldiag; i++)
   {
    fprintf(fd,"%d \t%lf %d \n",i,
                ldiag[i].Value,
                ldiag[i].PtrAix );
   } /* end for i */
 
printf("write out the last diagonal block row storage array of structures \n");
fflush(stdout);

/* (9) write out the last diagonal block row storage array of structures */
  fprintf(fd," \t%d\n",nlblockrow);
  for (i=1; i<=nlblockrow; i++)
   {
    fprintf(fd,"%d \t%lf %d %d \n",i,
                lblockrow[i].Value,
                lblockrow[i].i,
                lblockrow[i].j );
   } /* end for i */
 
printf("write out the matrix ulbbufsize[][] \n");
fflush(stdout);

/* (9) write out the matrix ulbbufsize[][] */
  for (i=1; i<=nprocs; i++)
    {
      fprintf(fd,"\t%d \t%d\n",i,nprocs);
      for (j=1; j<=nprocs; j++)
       {
        fprintf(fd,"%d %d\t%d \n",i,j,ulbbufsize[i][j]);
       } /* end for i */
    } /* end for i */

printf("write out a character string to identify EOF \n");
fflush(stdout);

/* (10) write out a character string to identify EOF */
  fprintf(fd,"\nEOF_input_BBDmatrix\n");

  fclose(fd);

  fd = fopen(FileNameb,"w");
  if (fd == NULL)
   {
    printf("can't open file %s\n",FileNameb);
    exit(1);
   } /* end if */
 
printf("write out the b vector \n");
fflush(stdout);
 
/* (11) write out the b vector */
  for (i=1; i<=n; i++)
   {
    fprintf(fd,"%lf\t",(2.0*drand48()));
    if( (i%5) == 0) fprintf(fd,"\n");
   } /* end for i */
 
/* (12) write out a character string to identify EOF */
  fprintf(fd,"\n\nEOF_input_b_vector\n");

 fclose(fd);
} /* end OutputBDBFormData() */


/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
void ColorLB()
{
  int i,j,node,color;
  int neighbors,coloredneighbors,MinColoredNodes;
  int ColoredNodes[10];
  Set *Neighbors;
  Set *ColoredNeighbors;
  Set *ColorsOfNeighbors;
  SetElement *ElementOfNeighbors;
  SetElement *ElementOfColoredNeighbors;
  SetElement *ElementOfColorsOfNeighbors;
  char *string1 = "Neighbors of Node";
  char *string2 = "Colored Neighbors of Node";
  char *string3 = "Colors of Neighbors of Node";
  char *name1,*name2,*name3;

/* initialize the nodes */
  for (i=1;i<=(N2->n);i++)
    {
      nodeLB[i].Color = 0;

      name1 = (char *) calloc(22,sizeof(char));
      if (name1 == NULL)
       {
        printf("can't allocate memory for name1[] \n");
        exit(1);
       } /* end if */
      for (j=0; j<17; j++)
        name1[j] = string1[j];
      name1[j++] = '-';
      name1[j++] = (char) (((i%1000)/100) + 48);
      name1[j++] = (char) (((i%100)/10) + 48);
      name1[j++] = (char) ((i%10) + 48);
      name1[j++] = '\0';
      Neighbors = NULL;
      Neighbors = InitSet(Neighbors,name1);
      nodeLB[i].Neighbors = Neighbors;
/*
PrintSet(nodeLB[i].Neighbors);
*/
      name2 = (char *) calloc(32,sizeof(char));
      if (name2 == NULL)
       {
        printf("can't allocate memory for name2[] \n");
        exit(1);
       } /* end if */

      for (j=0; j<25; j++)
        name2[j] = string2[j];
      name2[j++] = '-';
      name2[j++] = (char) (((i%1000)/100) + 48);
      name2[j++] = (char) (((i%100)/10) + 48);
      name2[j++] = (char) ((i%10) + 48);
      name2[j++] = '\0';
      ColoredNeighbors = NULL;
      ColoredNeighbors = InitSet(ColoredNeighbors,name2);
      nodeLB[i].ColoredNeighbors = ColoredNeighbors;
/*
PrintSet(nodeLB[i].ColoredNeighbors);
*/
      name3 = (char *) calloc(30,sizeof(char));
      if (name3 == NULL)
       {
        printf("can't allocate memory for name3[] \n");
        exit(1);
       } /* end if */

      for (j=0; j<27; j++)
        name3[j] = string3[j];
      name3[j++] = '-';
      name3[j++] = (char) (((i%1000)/100) + 48);
      name3[j++] = (char) (((i%100)/10) + 48);
      name3[j++] = (char) ((i%10) + 48);
      name3[j++] = '\0';
      ColorsOfNeighbors = NULL;
      ColorsOfNeighbors = InitSet(ColorsOfNeighbors,name3);
      nodeLB[i].ColorsOfNeighbors = ColorsOfNeighbors;
/*
PrintSet(nodeLB[i].ColorsOfNeighbors);
*/
    } /* end for i */
  free(name1);
  free(name2);
  free(name3);

/* initialize the histogram of insidences of colors */
  for (i=1;i<=10;i++)  ColoredNodes[i] = 0;

/* load data into nodelb[].Neighbors */
  for (i=1;i<=nadjlistLB;i++)
    {
/*
printf("adjlistLB[%d].l.NodeNumR=%d - adjlistLB[%d].r.NodeNumR=%d \n",
         i,adjlistLB[i].l.NodeNumR,i,adjlistLB[i].r.NodeNumR);
fflush(stdout);
*/
     AddElementToSet(nodeLB[adjlistLB[i].l.NodeNumR].Neighbors,adjlistLB[i].r.NodeNumR);
     AddElementToSet(nodeLB[adjlistLB[i].r.NodeNumR].Neighbors,adjlistLB[i].l.NodeNumR);
    } /* end for i */
/*
  for (i=1;i<=(N2->n);i++)  PrintSet(nodeLB[i].Neighbors);
*/
/* color the graph using a greedy algorithm based on the Saturation
   degree heuristic with the node with the highest valency being selected 
   as the first graph node to be colored */
  ncolors = 0;
  for (i=1;i<=(N2->n);i++)
    {
/* select the next node to be colored */
      node = 0;
      neighbors = -1;
      coloredneighbors = 0;
      for (j=1;j<=(N2->n);j++)
	{
         if ( ((nodeLB[j].ColoredNeighbors)->n > coloredneighbors) && (nodeLB[j].Color == 0) )
           {
            node = j;
            neighbors = (nodeLB[j].Neighbors)->n;
            coloredneighbors = (nodeLB[j].ColoredNeighbors)->n;
	   } /* end if */
	} /* end for j */
    if (node == 0)
       {
        for (j=1;j<=(N2->n);j++)
	  {
           if ( ((nodeLB[j].Neighbors)->n > neighbors) && (nodeLB[j].Color == 0) )
             {
              node = j;
              neighbors = (nodeLB[j].Neighbors)->n;
              coloredneighbors = (nodeLB[j].ColoredNeighbors)->n;
	     } /* end if */
	  } /* end for j */
       } /* end for node==0 */
/*
printf("i=%d - node=%d - neighbors=%d - coloredneighbors=%d \n",
        i,node,neighbors,coloredneighbors);
fflush(stdout);
*/
/* identify the colors of the neighbors and select a unique color 
   select the color that has been used the least frequent */
    if ( (nodeLB[node].ColorsOfNeighbors)->n == ncolors)
      {
       color = ++ncolors;
      } /* end if new color */
    else
      {
       color = 0;
       MinColoredNodes = (N2->n);
       for (j=1;j<=ncolors;j++)
         {
          if (ElementOfSet(nodeLB[node].ColorsOfNeighbors,j) == NO)
	    {
             if (ColoredNodes[j] < MinColoredNodes)
	       {
                color = j;
                MinColoredNodes = ColoredNodes[j];
	       } /* end if */
	    } /* end if */
         } /* end for j */
      } /* end if same color */
    nodeLB[node].Color = color;
    ColoredNodes[color]++;
/*
printf("\ti=%d - node=%d - color=%d - ColoredNodes[%d]=%d \n",
        i,node,color,color,ColoredNodes[color]);
fflush(stdout);
*/
/* update the color information in the neighbors */
    ElementOfNeighbors = (nodeLB[node].Neighbors)->First;
    while (ElementOfNeighbors != NULL)
     {
/*
printf("\t\ti=%d - node=%d - ElementOfNeighbors->element=%d \n",
        i,node,ElementOfNeighbors->Element);
fflush(stdout);
*/
      AddElementToSet(nodeLB[ElementOfNeighbors->Element].ColoredNeighbors,node);
      AddElementToSet(nodeLB[ElementOfNeighbors->Element].ColorsOfNeighbors,color);
/*
PrintSet(nodeLB[ElementOfNeighbors->Element].ColoredNeighbors);
PrintSet(nodeLB[ElementOfNeighbors->Element].ColorsOfNeighbors);
*/
      ElementOfNeighbors = ElementOfNeighbors->Next;
     } /* end while ElementOfNeighbors != NULL */
    } /* end for i */
} /* end ColorLB() */


/* ****************************************************************************** */
/*                                                                                */
/*  Function:  X-Image(image-index,size,size,title)                                     */
/*                                                                                */
/*  derived fron an X-window display program developed by M. L. Katz of the MITRE */
/*  Corporation.  The original program was called image256.c.                     */
/*                                                                                */
/*  Description:  This program takes a matrix containing bytes that               */
/*                represent a digital image and displays that image               */
/*                in an X11 window.                                               */
/*                                                                                */
/* ****************************************************************************** */

void Ximage(xgs,ygs,Title)
int xgs,ygs;
char *Title;
{
    Window win;
    int x = 0, y = 0;           /* window position */
    unsigned int border_width = 4;  /* border four pixels wide */
    char *window_name = Title;
    char *icon_name = "gray";
    Pixmap icon_pixmap;
    XSizeHints size_hints;
    XEvent report;
    GC gc;
    char *display_name = NULL;
    int window_size = 0;  /* OK, or too SMALL to display contents */
/*
printf("width=|%d| - height=|%d| - title=|%s| \n",xgs,ygs,Title);
fflush(stdout);	
*/
    /* 1) set window width and height as specified by the arguments */
    width  = xgs;
    height = ygs;
/*
printf("connect to the server \n");
fflush(stdout);
*/
    /* 2) connect to the server */
    if ( (display = XOpenDisplay(display_name)) == NULL )
    {
    (void) fprintf( stderr,
        "gray scale: cannot connect to X server %s\n",
        XDisplayName(display_name));
    exit(-1);
    }
/*
printf("set the screen \n");
fflush(stdout);
*/
    /* 3) set the screen */
    /* get screen size from display structure macro */
    screen = DefaultScreen(display);
/*
printf("create a window \n");
fflush(stdout);
*/
    /* 4) create a window */
    win = XCreateSimpleWindow(display, RootWindow(display,screen),
        x, y, width, height, border_width,
        BlackPixel(display, screen),
        WhitePixel(display, screen));
/*
printf("load the image data from the image matrix \n");
fflush(stdout);
*/
    /* 10) load the image data from the image matrix */
    load_data();
/*
printf("prepare an icon pixmap \n");
fflush(stdout);
*/
    /* 5) prepare an icon pixmap */
    icon_pixmap = XCreateBitmapFromData(display, win, icon_bitmap_bits,
    icon_bitmap_width, icon_bitmap_height);

    /* initialize size hint property for window manager */
    size_hints.flags = PPosition | PSize | PMinSize;
    size_hints.x = x;
    size_hints.y = y;
    size_hints.width = width;
    size_hints.height = height;
    size_hints.min_width = width;
    size_hints.min_height = height;
/*
printf("set standard window properties \n");
fflush(stdout);
*/
    /* 6) set standard window properties */
    XSetStandardProperties(display, win, window_name, icon_name,
    icon_pixmap, gargv, gargc, &size_hints);
/*
printf("select event types \n");
fflush(stdout);
*/
    /* 7) select event types */
    XSelectInput(display, win, ExposureMask | ButtonPressMask |
    StructureNotifyMask);
/*
printf("define a gray scale colormap \n");
fflush(stdout);
*/
    /* 8) define a gray scale colormap */
    set_gray_scale(win);
/*
printf("create a graphics context \n");
fflush(stdout);
*/
    /* 9) create a graphics context */
    get_GC(win, &gc);
/*
printf("display the window \n");
fflush(stdout);
*/
    /* 11) display the window */
    XMapWindow(display, win);
/*
printf("draw the image in the window\n");
fflush(stdout);
*/
    /* 12) draw the image in the window */
    XPutImage(display, win, gc, picture, 0, 0, 0, 0, width, height);
/*
printf("process events \n");
fflush(stdout);
*/
    /* 13) process events */
    /* get events, use first Expose to display graphics;
     * ConfigureNotify to iniate a resize;  ButtonPress to exit
     */
    while(1)
    {
    XNextEvent(display, &report);
    switch (report.type)
    {
        case Expose:
        /* get rid of all other Expose events on the queue */
        while (XCheckTypedEvent(display, Expose, &report));
        if (window_size == SMALL)
            /* if window was resized too small to use */
           break;
        else
            /* place image in window */
            XPutImage(display, win, gc, picture, 0, 0, 0, 0, width, height);
        break;
        case ConfigureNotify:
        /* window has been resized, change width and
         * height to send to draw_gray_scale in next
         * Expose
         */
        width = report.xconfigure.width;
        height = report.xconfigure.height;
        if ((width < size_hints.min_width) ||
            (height < size_hints.min_height))
            window_size = SMALL;
        else
            window_size = OK;
        break;
        case ButtonPress:
        XFreeGC(display, gc);
        XCloseDisplay(display);
        return;
        default:
        /* all events selected by StructureNotifyMask
         * except ConfigureNotify are thrown away here.
         * since nothing is done with them.
         */
        break;
    }  /* end switch */
    }  /* end while */
}

/* ***************************************************************** */
/*                                                                   */
/*  This routine creates and installs a colormap with 256 grayscale  */
/*  colors.                                                          */
/*                                                                   */
/* ***************************************************************** */
void set_gray_scale(win)
Window win;
{
    Visual *visual;
    Colormap cmap;
    int  i;
    XColor cell_defs[MAX_COLORS];
    int white, black;

    /* set visual to default visual */
    visual = DefaultVisual(display, screen);

    /* create colormap with default visual and all color cells allocated */
    cmap = XCreateColormap(display, win, visual, AllocAll);

    /* install the new colormap for the given window */
    XSetWindowColormap(display, win, cmap);

    black = BlackPixel(display, screen);
    white = WhitePixel(display, screen);

    /* set white pixel as in standard colormap
    cell_defs[255].pixel = white;                         white
    cell_defs[255].red = 65535;
    cell_defs[255].green = 65535;
    cell_defs[255].blue = 65535;
    cell_defs[255].flags = DoRed | DoGreen | DoBlue;

     set black pixel as in standard colormap
    cell_defs[0].pixel = black;                         black
    cell_defs[0].red = 0;
    cell_defs[0].green = 0;
    cell_defs[0].blue = 0;
    cell_defs[0].flags = DoRed | DoGreen | DoBlue;
*/
    /* set color cells 2-254 to represent gray scale colors */
/*    for (i=1; i<(MAX_COLORS-2); i++) */
    for (i=0; i<(MAX_COLORS); i++)
    {
    cell_defs[i].pixel = i;
    cell_defs[i].red = (i*65535)/(MAX_COLORS-1);
    cell_defs[i].green = (i*65535)/(MAX_COLORS-1);
    cell_defs[i].blue = (i*65535)/(MAX_COLORS-1);
    cell_defs[i].flags = DoRed | DoGreen | DoBlue;
    }

    /* store color cell colors in new colormap */
    XStoreColors(display, cmap, cell_defs, MAX_COLORS);
}

/* ******************************************************************* */
/*                                                                     */
/*  This routine creates a graphics context by calling the X Windows   */
/*  routine XCreateGC.  A default graphics context is created without  */
/*  specifying any of the attribute values.                            */
/*                                                                     */
/* ******************************************************************* */
void get_GC(win, gc)
Window win;
GC *gc;
{
    unsigned long valuemask = 0;  /* ignore XGCvalues and use defaults */
    XGCValues values;

    *gc = XCreateGC(display, win, valuemask, &values);
}


/* ***************************************************************** */
/*                                                                   */
/*  This routine loads the data file that represents the digital     */
/*  image.                                                           */
/*                                                                   */
/* ***************************************************************** */
int load_data()
{
    int i, j;
    unsigned char *pic;
    Visual *vis;
    int depth;

    /* allocate memory to store image */
    if ((pic = (unsigned char *) calloc(width, height, sizeof(unsigned char)))
                            == NULL)
    {
    sprintf(stderr, "Error allocating array\n");
    return(0);
    }
    /* adjust values in image to correspond to colormap */
      for (i = 0; i < height; i++)
    for (j = 0; j < width; j++)
          {
            *(pic+(i*width+j)) = bitmap[j+1][i+1];

        if (*(pic+(i*width+j)) == 255)
        *(pic+(i*width+j)) = 0;
        else
        (*(pic+(i*width+j)))++;
    }

    vis = DefaultVisual(display, screen);
    depth = DisplayPlanes(display, screen);

    /* create an image structure in X */
    picture = XCreateImage(display, vis, depth, ZPixmap, 0,
        pic, width, height, 8, width);
}


